# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to use lintian in debusine."""

import json
import re
from collections import defaultdict
from pathlib import Path
from typing import Any, override

from debian import debfile

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine import utils
from debusine.artifacts import LintianArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianLintianSummary,
    DebianSourcePackage,
    DebianSystemImage,
    DebianSystemTarball,
    DebianUpload,
    WorkRequestResults,
    get_source_package_name,
    get_source_package_version,
)
from debusine.client.models import RelationType
from debusine.tasks import BaseTaskWithExecutor, RunCommandTask
from debusine.tasks.inputs import (
    MultiInput,
    OptionalEnvironmentInput,
    OptionalSingleInput,
    Stage,
)
from debusine.tasks.models import LintianData, LintianDynamicData


class Lintian(
    RunCommandTask[LintianData, LintianDynamicData],
    BaseTaskWithExecutor[LintianData, LintianDynamicData],
):
    """Task to use lintian in debusine."""

    TASK_VERSION = 1

    # Keep _SEVERITY_SHORT_TO_LONG ordered from higher to lower
    _SEVERITY_SHORT_TO_LONG = {
        "E": "error",
        "W": "warning",
        "I": "info",
        "P": "pedantic",
        "X": "experimental",
        "O": "overridden",
        "C": "classification",  # ignored for the fail_on_severity
        "M": "masked",  # In _IGNORED_SEVERITIES: not exposed to the user
    }

    # Severities that are not exposed to the user
    _IGNORED_SEVERITIES = {"masked"}

    CAPTURE_OUTPUT_FILENAME = "lintian.txt"

    input_source_artifact = OptionalSingleInput(
        field="input.source_artifact",
        categories=[ArtifactCategory.SOURCE_PACKAGE, ArtifactCategory.UPLOAD],
        stage=Stage.PENDING,
    )
    input_binary_artifacts = MultiInput(
        field="input.binary_artifacts",
        categories=[ArtifactCategory.BINARY_PACKAGE, ArtifactCategory.UPLOAD],
        stage=Stage.PENDING,
    )
    environment = OptionalEnvironmentInput(
        categories=[
            ArtifactCategory.SYSTEM_TARBALL,
            ArtifactCategory.SYSTEM_IMAGE,
        ],
        stage=Stage.PENDING,
    )

    @override
    def post_init(self) -> None:
        super().post_init()

        # In the lintian cmd: the target that is tested (local file, .dsc
        # or .changes)
        self._lintian_targets: list[Path] | None = None

        # Architecture to package names that are being analysed by this
        # Lintian invocation. In other words: packages belonging to each
        # architecture.
        self._architecture_to_packages: defaultdict[str, set[str]] = (
            defaultdict(set)
        )

        self._package_name_to_filename: dict[str, str] = {}

        self._severities_to_level = {}  # dictionary with severity's name to
        # level. Higher is bigger severity
        for ranking, severity in enumerate(
            reversed(self._SEVERITY_SHORT_TO_LONG.values())
        ):
            self._severities_to_level[severity] = ranking

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        if self.input_source_artifact is not None:
            assert isinstance(
                self.input_source_artifact.data,
                (DebianSourcePackage, DebianUpload),
            )
            provided.add(
                ttags.ProvenanceProvided.SYSTEM,
                {
                    ttags.SOURCE_PACKAGE_PREFIX
                    + get_source_package_name(self.input_source_artifact.data)
                },
            )

        return provided, required

    @override
    def get_subject(self) -> str | None:
        if self.input_source_artifact is not None:
            assert isinstance(
                self.input_source_artifact.data,
                (DebianSourcePackage, DebianUpload),
            )
            return get_source_package_name(self.input_source_artifact.data)

        for binary_artifact in self.input_binary_artifacts.artifacts:
            assert isinstance(
                binary_artifact.data,
                (DebianBinaryPackage, DebianUpload),
            )
            return get_source_package_name(binary_artifact.data)

        return None

    @override
    def get_configuration_context(self) -> str | None:
        if (environment := self.environment) is not None:
            assert isinstance(
                environment.data, (DebianSystemTarball, DebianSystemImage)
            )
            return environment.data.codename
        return None

    @override
    def compute_dynamic_data(self) -> LintianDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: name of source package of ``input.source_artifact`` or
          ``input.binary_artifacts``
        :parameter_summary: source package name and Version
        :runtime_context: ``what:distribution`` where ``what`` is what
          got analyzed (``source``, ``binary-all`` or ``binary-any``) and
          ``distribution`` is ``codename`` of ``environment``.
        :configuration_context: ``codename`` of ``environment``
        """
        version = None
        input_source_artifact = self.input_source_artifact
        if input_source_artifact is not None:
            assert isinstance(
                input_source_artifact.data, (DebianSourcePackage, DebianUpload)
            )
            version = get_source_package_version(input_source_artifact.data)

        # Validate input_binary_artifacts and get subject
        for binary_artifact in self.input_binary_artifacts.artifacts:
            assert isinstance(
                binary_artifact.data,
                (DebianBinaryPackage, DebianUpload),
            )
            if version is None:
                version = get_source_package_version(binary_artifact.data)

        what = {"source", "binary-all", "binary-any"}

        output = self.data.output
        if not output.source_analysis:
            what.remove("source")
        if not output.binary_all_analysis:
            what.remove("binary-all")
        if not output.binary_any_analysis:
            what.remove("binary-any")

        what_str = "+".join(sorted(what))

        runtime_context = None
        environment_id = None
        if (environment := self.environment) is not None:
            environment_id = environment.artifact_id
            assert isinstance(
                environment.data, (DebianSystemTarball, DebianSystemImage)
            )

            runtime_context = f"{what_str}:{environment.data.codename}"

        subject = self.get_subject()
        return LintianDynamicData(
            environment_id=environment_id,
            input_source_artifact_id=(
                None
                if input_source_artifact is None
                else input_source_artifact.artifact_id
            ),
            input_binary_artifacts_ids=self.input_binary_artifacts.resolved_ids(),
            subject=subject,
            parameter_summary=f"{subject}_{version}",
            runtime_context=runtime_context,
            configuration_context=self.get_configuration_context(),
        )

    @classmethod
    def generate_severity_count_zero(cls) -> dict[str, int]:
        """Return dictionary. Keys: severities. Value: 0."""
        return {
            value: 0
            for value in cls._SEVERITY_SHORT_TO_LONG.values()
            if value not in cls._IGNORED_SEVERITIES
        }

    def _create_analysis(
        self,
        parsed_output: list[dict[str, str]],
        packages: set[str],
    ) -> dict[str, Any]:
        """
        Parse output of lintian, return the analysis.

        :param parsed_output: output of lintian (parsed by _parse_output())
        :param packages: package names to consider for this analysis
        :return: Dictionary with "tags" key and the list of tags (sorted by
          package, severity, tag and tag_information)
        """
        # Create a copy of each tag, remove " source" from the package name
        # (analysis.json does not have " source", but "parsed_output" is
        # the output from `lintian` and has it

        # Create dictionary with all severities and 0 in the value
        tags_count_by_severity = self.generate_severity_count_zero()

        tags_found = set()
        overridden_tags_found = set()
        package_filename = {}
        tags_for_packages = []

        for tag in parsed_output:
            package = tag["package"]

            if package not in packages:
                continue

            package_no_source_suffix = package.removesuffix(" source")
            package_filename[package_no_source_suffix] = (
                self._package_name_to_filename[package]
            )

            tags_for_packages.append(
                {**tag, "package": package_no_source_suffix}
            )

            tag_name = tag["tag"]

            tags_count_by_severity[tag["severity"]] += 1

            if tag["severity"] == "overridden":
                overridden_tags_found.add(tag_name)
            else:
                tags_found.add(tag_name)

        summary = {
            "tags_count_by_severity": tags_count_by_severity,
            "package_filename": package_filename,
            "tags_found": sorted(tags_found),
            "overridden_tags_found": sorted(overridden_tags_found),
            "lintian_version": self._get_lintian_version(),
            "distribution": self.data.target_distribution,
        }

        return {
            "tags": tags_for_packages,
            "summary": summary,
            "version": self.TASK_VERSION,
        }

    @classmethod
    def _parse_tag_line(cls, line: str) -> dict[str, str] | None:
        """
        Return dictionary with the tag information or None.

        Lintian in Debian bullseye return:
        W: hello source: package-uses-deprecated-debhelper-compat-version 9
        N:
        W: package-uses-deprecated-debhelper-compat-version

        The second "W:" line is a tag_line but _parse_tag_line() return
        None: the information is already included. This line can be ignored.
        """
        severities = "".join(cls._SEVERITY_SHORT_TO_LONG)

        m = re.match(
            rf"(?P<severity>[{severities}M]):\s(?P<package>\S+"
            rf"(?: source)?):\s(?P<tag>\S+) ?(?P<note>[^[]*)?",
            line,
        )
        # severity, package and tags are always there. Information might
        # or might not be there in line. If it's there is does NOT start
        # with [ (if it starts with [ it's a file, dealt below)

        if m is None:
            m = re.match(rf"(?P<severity>[{severities}M]):\s(?P<tag>\S+)", line)
            if m:
                # Bullseye's lintian have a duplicated tag line after each
                # real (full information). Ignore it.
                # Example:
                #
                # W: hello source: package-uses-deprecated-debhelper-compat-v 9
                # N:
                # W: package-uses-deprecated-debhelper-compat-v
                #
                # The second W: line is ignored.
                return None

            raise ValueError(f"Failed to parse line: {line}")

        parsed_tag = {
            "severity": cls._SEVERITY_SHORT_TO_LONG[m.group("severity")],
            "package": m.group("package"),
            "tag": m.group("tag"),
            "explanation": "",
            "comment": "",
            "note": (m.group("note") or "").strip(),
            "pointer": "",
        }

        m = re.match(r".*\[(?P<pointer>.*)]", line)

        if m:
            parsed_tag["pointer"] = m.group("pointer")

        return parsed_tag

    @classmethod
    def _parse_output(cls, lintian_output: Path) -> list[dict[str, str]]:
        """
        Parse output of lintian, return structured representation.

        :param lintian_output: file with lintian's output
        :return: list with a dictionary per each lintian line. Keys:
            severity, package, tag, details (if it's in the output), file
            (if it's in the output)
        """
        tags = []

        tag: dict[str, str] | None = None
        previous_line: str | None = None

        in_comment = False
        explanation: str = ""
        comment: str = ""

        with lintian_output.open() as output:
            for line in output:
                line = line.strip()
                if line == "N:" and previous_line == "N:":
                    # Comments (written by the maintainer in the lintian's
                    # override files) and explanations (all start by N:).
                    # The comments starts when there are "N:\nN:\n" in the
                    # output.
                    in_comment = True

                elif line.startswith("N:"):
                    parsed_line = line.lstrip("N:").lstrip() + "\n"
                    if in_comment:
                        comment += parsed_line
                    elif explanation != "" or parsed_line != "\n":
                        explanation += parsed_line

                else:
                    # Line not starting with "N:". It is a new tag.

                    # Add the explanation into the current tag before
                    # creating the new one
                    if tag:
                        tag["explanation"] = explanation.strip()
                        explanation = ""

                    if (new_tag := cls._parse_tag_line(line)) is not None:
                        tag = new_tag

                        # If a comment was read (above the tag) add it
                        tag["comment"] = comment.strip()
                        comment = ""
                        # If we were in a comment: not anymore
                        in_comment = False

                        # Add the tag in the list, unless it is needed
                        # to be ignored (e.g. "masked")
                        if tag["severity"] not in cls._IGNORED_SEVERITIES:
                            tags.append(tag)

                previous_line = line

            # Add explanation for the last tag (usually "explanation"
            # is added when finding a new tag)
            if tag is not None:
                tag["explanation"] = explanation.strip()

        return tags

    @override
    def _cmdline(self) -> list[str]:
        """
        Build the lintian command line.

        Use configuration of self.data and self._lintian_targets.
        """
        cmd = [
            "lintian",
            "--no-cfg",
            "--display-level",
            ">=classification",
            "--display-experimental",
            "--info",
            "--show-overrides",
        ]

        if include_tags := self.data.include_tags:
            cmd.append("--tags=" + ",".join(include_tags))

        if exclude_tags := self.data.exclude_tags:
            cmd.append("--suppress-tags=" + ",".join(exclude_tags))

        # Set by configure_for_execution, which is always run before this.
        assert self._lintian_targets is not None

        cmd.extend([str(target) for target in self._lintian_targets])

        return cmd

    def _extract_package_name_arch(
        self, file: Path
    ) -> tuple[str, str] | tuple[None, None]:
        """
        Return package name and architecture for file.

        :param file: if file ends in .deb or .udeb: use debfile.DebFile
            to return control["Package"]. If it's a .dsc: return
            dsc["Source"] + source. For other file extensions: return None.
        :return: package filename.
        """
        if file.suffix in (".deb", ".udeb"):
            pkg = debfile.DebFile(file)
            control = pkg.control.debcontrol()
            package_name = control["Package"]
            architecture = control["Architecture"]

        elif file.suffix in ".dsc":
            try:
                dsc = utils.read_dsc(file)
            except ValueError as e:
                self.append_to_log_file("configure_for_execution.log", [str(e)])
                return None, None

            # Lintian identifies the binary package Vs. source package with
            # the string " source". This is removed when presented to the user
            package_name = dsc["Source"] + " source"
            architecture = "source"
        else:
            package_name = None
            architecture = None

        return package_name, architecture

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Download the required artifacts."""
        assert self.dynamic_data

        if (
            artifact_id := self.dynamic_data.input_source_artifact_id
        ) is not None:
            self.fetch_artifact(artifact_id, destination)

        for artifact_id in self.dynamic_data.input_binary_artifacts_ids:
            self.fetch_artifact(artifact_id, destination)

        return True

    @override
    def configure_for_execution(self, download_directory: Path) -> bool:
        """
        Find a .dsc, .deb and .udeb files in download_directory.

        Set self._lintian_targets to the relevant files.

        :param download_directory: where to search the files
        :return: True if valid files were found
        """
        self._prepare_executor_instance()

        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")

        self.run_executor_command(
            ["apt-get", "update"],
            log_filename="install.log",
            run_as_root=True,
            check=True,
        )
        self.run_executor_command(
            ["apt-get", "--yes", "install", "lintian"],
            log_filename="install.log",
            run_as_root=True,
            check=True,
        )

        self._lintian_targets = utils.find_files_suffixes(
            download_directory, [".dsc"]
        ) + utils.find_files_suffixes(download_directory, [".deb", ".udeb"])

        # Prepare mapping from filename to package name
        for file in download_directory.iterdir():
            package_name, architecture = self._extract_package_name_arch(
                download_directory / file
            )
            if package_name is None or architecture is None:
                continue

            self._package_name_to_filename[package_name] = file.name
            self._architecture_to_packages[architecture].add(package_name)

        if len(self._lintian_targets) == 0:
            ignored_file_names = [
                file.name for file in download_directory.iterdir()
            ]
            ignored_file_names.sort()

            self.append_to_log_file(
                "configure_for_execution.log",
                [
                    f"No *.dsc, *.deb or *.udeb to be analyzed. "
                    f"Files: {ignored_file_names}"
                ],
            )
            return False

        return True

    @override
    def check_directory_for_consistency_errors(
        self, build_directory: Path
    ) -> list[str]:
        """Return list of errors."""
        files_in_directory: set[Path] = set(build_directory.iterdir())

        if (
            build_directory / self.CAPTURE_OUTPUT_FILENAME
            not in files_in_directory
        ):
            return [f"{self.CAPTURE_OUTPUT_FILENAME} not in {build_directory}"]

        return []

    @override
    def upload_artifacts(
        self,
        exec_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Upload the LintianArtifact with the files and relationships."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        lintian_file = exec_directory / self.CAPTURE_OUTPUT_FILENAME

        architecture_to_analysis = self._create_architecture_to_analysis(
            lintian_file
        )

        for architecture, analysis in architecture_to_analysis.items():
            (
                analysis_file := exec_directory
                / f"analysis-{architecture}.json"
            ).write_text(json.dumps(analysis))

            lintian_artifact = LintianArtifact.create(
                analysis=analysis_file,
                lintian_output=lintian_file,
                architecture=architecture,
                summary=DebianLintianSummary(**analysis["summary"]),
            )

            uploaded = self.debusine.upload_artifact(
                lintian_artifact,
                workspace=self.workspace_name,
                work_request=self.work_request_id,
            )

            for source_artifact_id in self._source_artifacts_ids:
                self.debusine.relation_create(
                    uploaded.id, source_artifact_id, RelationType.RELATES_TO
                )

    def _create_architecture_to_analysis(
        self, lintian_file: Path
    ) -> dict[str, dict[str, Any]]:
        """
        Read lintian_file, return dict mapping architecture to analysis.

        :param lintian_file: Lintian output file
        :return: dict[str, Analysis]:
          only for the architectures selected by self.data.output
        """
        parsed_output = self._parse_output(lintian_file)

        parsed_output.sort(
            key=lambda x: (
                x["package"],
                -self._severities_to_level[
                    x["severity"]
                ],  # negative to sort from highest to lowest
                x["tag"],
                x["note"],
            )
        )

        package_analysis_mapping = defaultdict(lambda: "binary_any_analysis")
        package_analysis_mapping["source"] = "source_analysis"
        package_analysis_mapping["all"] = "binary_all_analysis"
        output = self.data.output

        architecture_to_analysis: dict[str, dict[str, Any]] = {}

        for architecture, packages in self._architecture_to_packages.items():
            if not getattr(output, package_analysis_mapping[architecture]):
                # self.data.output.source_analysis (or binary_all_analysis,
                # or binary_any_analysis) was disabled
                continue

            architecture_to_analysis[architecture] = self._create_analysis(
                parsed_output, packages
            )

        return architecture_to_analysis

    @override
    def task_result(
        self, returncode: int | None, execute_directory: Path
    ) -> WorkRequestResults:
        """
        Evaluate task output and return success.

        For a successful run of lintian:
        - exit code must be 0 (normal operation) or 2 (detected a condition
          specified via the ``--fail-on`` option, which defaults to
          ``error``)
        - lintian must have generated the output file
        - no tags of severity ``self.data["fail_on"]`` or higher

        :return: SUCCESS for success, FAILURE for failure due to tags of
          severity ``self.data["fail_on"]`` or higher, or ERROR if the exit
          code was not 0 or 2.
        """
        if returncode not in {0, 2}:
            return WorkRequestResults.ERROR

        fail_on_severity = self.data.fail_on_severity

        if fail_on_severity == "none":
            return WorkRequestResults.SUCCESS

        fail_on_severity_level = self._severities_to_level[fail_on_severity]
        lintian_file = execute_directory / self.CAPTURE_OUTPUT_FILENAME
        parsed_output = self._parse_output(lintian_file)

        for tag in parsed_output:
            if (
                self._severities_to_level[tag["severity"]]
                >= fail_on_severity_level
            ):
                return WorkRequestResults.FAILURE

        return WorkRequestResults.SUCCESS

    def _get_lintian_version(self) -> str:
        if not self.executor_instance:
            raise AssertionError(
                "Task.executor_instance must be set before calling "
                "_get_lintian_version()"
            )

        out = self.executor_instance.run(
            ["lintian", "--print-version"],
            encoding="utf-8",
            errors="ignore",
            text=True,
            check=True,
        ).stdout

        return out.strip()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Database models for the Debusine signing service."""

from collections.abc import Sequence
from pathlib import Path
from typing import Any, BinaryIO, ClassVar, TYPE_CHECKING, TypeAlias

from django.conf import settings
from django.db import models

from debusine.assets import KeyPurpose
from debusine.signing.models import ProtectedKey, ProtectedKeyNaCl, SigningMode
from debusine.signing.openssl import openssl_generate, x509_fingerprint
from debusine.signing.sbsign import sbsign
from debusine.utils import calculate_hash

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer
    from django_stubs_ext.db.models import TypedModelMeta

    ReadableBuffer  # fake usage for vulture
else:
    TypedModelMeta = object


class AuditLog(models.Model):
    """An audit log entry."""

    objects: ClassVar[models.Manager["AuditLog"]] = models.Manager["AuditLog"]()

    class Event(models.TextChoices):
        GENERATE = "generate", "Generate"
        SIGN = "sign", "Sign"
        REGISTER = "register", "Register an externally-generated key"

    purpose = models.CharField(max_length=32, choices=KeyPurpose.choices)
    fingerprint = models.CharField(max_length=64)
    event = models.CharField(max_length=8, choices=Event.choices)
    data = models.JSONField()
    created_at = models.DateTimeField(auto_now_add=True)
    created_by_work_request_id = models.IntegerField(blank=True, null=True)

    class Meta(TypedModelMeta):
        app_label = "signing"
        indexes = [
            models.Index(
                "purpose",
                "fingerprint",
                name="%(app_label)s_%(class)s_key_idx",
            ),
            models.Index(
                "created_by_work_request_id",
                name="%(app_label)s_%(class)s_wr_idx",
            ),
        ]


class KeyManager(models.Manager["Key"]):
    """Manager for the Key model."""

    def get_fingerprint(
        self,
        *,
        purpose: KeyPurpose,
        public_key: "ReadableBuffer",
        log_file: BinaryIO | None = None,
    ) -> str:
        """Get the fingerprint for a public key."""
        match purpose:
            case KeyPurpose.UEFI:
                return x509_fingerprint(public_key, log_file=log_file)
            case KeyPurpose.OPENPGP_UPLOAD | KeyPurpose.OPENPGP_REPOSITORY:
                # Imported late because the gpg module has to be installed as a
                # .deb.
                from debusine.signing.gnupg import gpg_fingerprint

                return gpg_fingerprint(public_key)
            case _ as unreachable:
                raise AssertionError(f"Unexpected purpose: {unreachable}")

    def generate(
        self,
        purpose: KeyPurpose,
        description: str,
        created_by_work_request_id: int,
        log_file: BinaryIO | None = None,
    ) -> "Key":
        """Generate a new key."""
        match purpose:
            case KeyPurpose.UEFI:
                # For now we just hardcode the certificate lifetime: 15
                # years is enough to extend from the start of the
                # development period of a Debian release to the end of
                # Freexian's extended LTS period.
                days = 15 * 365
                private_key, public_key = openssl_generate(
                    description, days, log_file=log_file
                )
            case KeyPurpose.OPENPGP_UPLOAD | KeyPurpose.OPENPGP_REPOSITORY:
                # Imported late because the gpg module has to be installed as a
                # .deb.
                from debusine.signing.gnupg import gpg_generate

                private_key, public_key = gpg_generate(description)
            case _ as unreachable:
                raise AssertionError(f"Unexpected purpose: {unreachable}")

        fingerprint = self.get_fingerprint(
            purpose=purpose, public_key=public_key, log_file=log_file
        )
        key = self.create(
            purpose=purpose,
            fingerprint=fingerprint,
            private_key=ProtectedKeyNaCl.encrypt(
                settings.DEBUSINE_SIGNING_PRIVATE_KEYS[0].public_key,
                private_key,
            ).model_dump(),
            public_key=public_key,
        )
        AuditLog.objects.create(
            event=AuditLog.Event.GENERATE,
            purpose=purpose,
            fingerprint=fingerprint,
            data={"description": description},
            created_by_work_request_id=created_by_work_request_id,
        )
        return key


class Key(models.Model):
    """A key managed by the debusine signing service."""

    Purpose: TypeAlias = KeyPurpose

    purpose = models.CharField(max_length=32, choices=Purpose.choices)
    fingerprint = models.CharField(max_length=64)
    private_key = models.JSONField()
    public_key = models.BinaryField()
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(blank=True, null=True)

    objects = KeyManager()

    class Meta(TypedModelMeta):
        app_label = "signing"
        constraints = [
            models.UniqueConstraint(
                fields=("purpose", "fingerprint"),
                name="%(app_label)s_%(class)s_unique_purpose_fingerprint",
            )
        ]

    @property
    def stored_private_key(self) -> ProtectedKey:
        """Access private_key as a pydantic model."""
        return ProtectedKey.model_validate(self.private_key)


def sign(
    keys: Sequence["Key"],
    data_path: Path,
    signature_path: Path,
    mode: SigningMode,
    created_by_work_request_id: int,
    username: str | None = None,
    user_id: int | None = None,
    resource: dict[str, Any] | None = None,
    log_file: BinaryIO | None = None,
) -> None:
    """
    Sign data in `data_path` using `keys`.

    `signature_path` will be overwritten with the resulting signature.
    """
    purposes = list({key.purpose for key in keys})
    assert len(purposes) == 1
    [purpose] = purposes

    match (purpose, mode):
        case KeyPurpose.UEFI, SigningMode.ATTACHED | SigningMode.DETACHED:
            assert len(keys) == 1
            sbsign(
                keys[0].stored_private_key,
                keys[0].public_key,
                data_path=data_path,
                signature_path=signature_path,
                detached=(mode == SigningMode.DETACHED),
                log_file=log_file,
            )
        case (
            KeyPurpose.OPENPGP_UPLOAD | KeyPurpose.OPENPGP_REPOSITORY,
            SigningMode.DETACHED | SigningMode.CLEAR,
        ):
            # Imported late because the gpg module has to be installed as a
            # .deb.
            from debusine.signing.gnupg import gpg_sign

            gpg_sign(
                [(key.stored_private_key, key.public_key) for key in keys],
                data_path=data_path,
                signature_path=signature_path,
                mode=mode,
            )
        case KeyPurpose.OPENPGP_UPLOAD, SigningMode.DEBSIGN:
            # Imported late because the gpg module has to be installed as a
            # .deb.
            from debusine.signing.gnupg import gpg_debsign

            assert len(keys) == 1
            gpg_debsign(
                keys[0].stored_private_key,
                keys[0].public_key,
                keys[0].fingerprint,
                data_path=data_path,
                signature_path=signature_path,
                log_file=log_file,
            )
        case _ as unreachable:
            raise AssertionError(f"Unexpected purpose and mode: {unreachable}")

    audit_log_data: dict[str, Any] = {
        "data_sha256": calculate_hash(data_path, "sha256").hex(),
    }
    if resource:
        audit_log_data["resource"] = resource
    if username:
        audit_log_data["username"] = username
    if user_id:
        audit_log_data["user_id"] = user_id

    AuditLog.objects.bulk_create(
        [
            AuditLog(
                event=AuditLog.Event.SIGN,
                purpose=purpose,
                fingerprint=key.fingerprint,
                data=audit_log_data,
                created_by_work_request_id=created_by_work_request_id,
            )
            for key in keys
        ]
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test the Debusine extensions to Signon."""

import contextlib
from collections.abc import Generator
from typing import Any, ClassVar, override
from unittest import mock

import responses
from django.test import RequestFactory, override_settings
from rest_framework import status

from debusine.db.models import Group, Identity, Scope
from debusine.server.signon import providers
from debusine.server.signon.sites import DebianSignon
from debusine.server.signon.tests.test_signon import SignonTestCase


@contextlib.contextmanager
def nm_status(status: str | None = None) -> Generator[None]:
    """
    Mock fetch_nm_claims.

    :param status: status to return from nm, or None for empty claims
    """
    return_value = {}
    if status is not None:
        return_value["nm_status"] = status
    with mock.patch(
        "debusine.server.signon.sites.DebianSignon.fetch_nm_claims",
        return_value=return_value,
    ) as fetch_nm_claims:
        yield
    fetch_nm_claims.assert_called()


@contextlib.contextmanager
def nm_not_called() -> Generator[None]:
    """Ensure fetch_nm_claims is not called."""
    with mock.patch(
        "debusine.server.signon.sites.DebianSignon.fetch_nm_claims",
        side_effect=AssertionError("fetch_nm_claims unexpectedly called"),
    ):
        yield


class TestDebianSignon(SignonTestCase):
    """Test DebianSignon."""

    scope: ClassVar[Scope]
    debian_group: ClassVar[Group]
    dm_group: ClassVar[Group]
    elts_group: ClassVar[Group]
    identity: ClassVar[Identity]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        debian_scope = cls.playground.get_or_create_scope("debian")
        freexian_scope = cls.playground.get_or_create_scope("freexian")
        cls.scope = cls.playground.get_default_scope()
        cls.debian_group = cls.playground.create_group(
            name="Debian", scope=debian_scope
        )
        cls.dm_group = cls.playground.create_group(
            name="DM", scope=debian_scope
        )
        cls.elts_group = cls.playground.create_group(
            name="elts-contributors", scope=freexian_scope
        )
        cls.identity = Identity.objects.create(
            issuer="test",
            subject="123",
            claims={
                "sub": 123,
                "name": "Test User",
                "email": "test@debian.org",
                "email_verified": True,
            },
        )

    @override
    def setUp(self) -> None:
        """Provide a mock unauthenticated request for tests."""
        super().setUp()
        self.factory = RequestFactory()
        self.request = self.factory.get("/")

    def setup_provider(self, name: str, add_to_group: Any = None) -> None:
        self.provider = self.make_gitlab_provider(
            name=name, add_to_group=add_to_group
        )
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        self.identity.issuer = name
        self.identity.save()

    @nm_not_called()
    def test_validate_claims_non_gitlab(self) -> None:
        provider = providers.Provider(name="unknown", label="Unknown")
        self.enterContext(override_settings(SIGNON_PROVIDERS=[provider]))
        signon = DebianSignon(self.request)
        self.assertEqual(
            signon.validate_claims(provider, self.identity),
            [
                "unsupported validation of logins from 'unknown' provider:"
                " Provider is not supported"
            ],
        )

    @nm_not_called()
    def test_validate_claims_email_not_verified(self) -> None:
        self.identity.claims["email_verified"] = False
        self.identity.save()
        self.setup_provider("unknown")
        signon = DebianSignon(self.request)
        self.assertEqual(
            signon.validate_claims(self.provider, self.identity),
            [f"identity {self.identity} does not have a verified email"],
        )

    @nm_status()
    def test_validate_claims_salsa_not_dd(self) -> None:
        self.setup_provider("salsa")
        signon = DebianSignon(self.request)
        self.assertEqual(
            signon.validate_claims(self.provider, self.identity),
            ["identity salsa:123 is not DD or DM"],
        )
        self.identity.refresh_from_db()
        self.assertNotIn("nm_status", self.identity.claims)

    @nm_status("dd")
    def test_validate_claims_salsa_dd(self) -> None:
        self.identity.claims["groups_direct"] = ["debian"]
        self.identity.save()
        self.setup_provider("salsa")
        signon = DebianSignon(self.request)
        self.assertEqual(
            signon.validate_claims(self.provider, self.identity), []
        )
        self.identity.refresh_from_db()
        self.assertEqual(self.identity.claims["nm_status"], "dd")

    @nm_status("dm")
    def test_validate_claims_salsa_dm(self) -> None:
        self.setup_provider("salsa")
        signon = DebianSignon(self.request)
        self.assertEqual(
            signon.validate_claims(self.provider, self.identity), []
        )
        self.identity.refresh_from_db()
        self.assertEqual(self.identity.claims["nm_status"], "dm")

    @nm_status("dm_ga")
    def test_validate_claims_salsa_dm_ga(self) -> None:
        """Log in a DM with guest account."""
        self.setup_provider("salsa")
        signon = DebianSignon(self.request)
        self.assertEqual(
            signon.validate_claims(self.provider, self.identity), []
        )
        self.identity.refresh_from_db()
        self.assertEqual(self.identity.claims["nm_status"], "dm_ga")

    @nm_status("dd_r")
    def test_validate_claims_salsa_dd_r(self) -> None:
        """Removed DDs are not allowed."""
        self.setup_provider("salsa")
        signon = DebianSignon(self.request)
        self.assertEqual(
            signon.validate_claims(self.provider, self.identity),
            ["identity salsa:123 is not DD or DM"],
        )
        self.identity.refresh_from_db()
        self.assertEqual(self.identity.claims["nm_status"], "dd_r")

    @nm_not_called()
    def test_create_user_add_to_group_legacy_string(self) -> None:
        self.setup_provider("test", add_to_group="debian/Debian")
        signon = DebianSignon(self.request)
        salsa_groups: None | list[str]
        for salsa_groups in (
            None,
            [],
            ["debian"],
            ["freexian/teams/lts-contributors"],
            ["debian", "freexian/teams/lts-contributors"],
        ):
            with self.subTest(salsa_groups=salsa_groups):
                if salsa_groups is None:
                    self.identity.claims.pop("groups_direct", None)
                else:
                    self.identity.claims["groups_direct"] = salsa_groups
                self.identity.save()
                user = signon.create_user_from_identity(self.identity)
                assert user is not None
                try:
                    self.assertQuerySetEqual(
                        user.debusine_groups.all(),
                        [],
                    )
                finally:
                    self.identity.user = None
                    self.identity.save()
                    user.delete()

    @nm_not_called()
    def test_create_user_add_to_group(self) -> None:
        self.setup_provider(
            "test",
            add_to_group={
                "debian": "debian/Debian",
                "freexian/teams/lts-contributors": "freexian/elts-contributors",
                "nm:dd": "debian/Debian",
                "nm:dm": "debian/DM",
                "nm:dm_ga": "debian/DM",
            },
        )
        signon = DebianSignon(self.request)
        salsa_groups: None | list[str]
        for salsa_groups, nm_status, debusine_groups in (
            (None, None, []),
            ([], None, []),
            (["debian"], None, [self.debian_group]),
            (["freexian/teams/lts-contributors"], None, [self.elts_group]),
            (
                ["debian", "freexian/teams/lts-contributors"],
                None,
                [self.debian_group, self.elts_group],
            ),
            ([], "dd", [self.debian_group]),
            ([], "dm", [self.dm_group]),
            ([], "dm_ga", [self.dm_group]),
            ([], "dd_r", []),
            (["debian"], "dd", [self.debian_group]),
        ):
            with self.subTest(salsa_groups=salsa_groups, nm_status=nm_status):
                if salsa_groups is None:
                    self.identity.claims.pop("groups_direct", None)
                else:
                    self.identity.claims["groups_direct"] = salsa_groups
                if nm_status is not None:
                    self.identity.claims["nm_status"] = nm_status
                self.identity.save()
                user = signon.create_user_from_identity(self.identity)
                assert user is not None
                signon.add_user_to_groups(
                    user, self.identity, self.provider.options["add_to_group"]
                )
                try:
                    self.assertQuerySetEqual(
                        user.debusine_groups.all(),
                        debusine_groups,
                        ordered=False,
                    )
                finally:
                    self.identity.user = None
                    self.identity.save()
                    user.delete()

    @nm_not_called()
    def test_create_user_unknown_provider(self) -> None:
        provider = providers.Provider(name="unknown", label="Unknown")
        self.enterContext(override_settings(SIGNON_PROVIDERS=[provider]))
        signon = DebianSignon(self.request)
        user = signon.create_user_from_identity(self.identity)
        assert user is not None
        self.assertQuerySetEqual(user.debusine_groups.all(), [])

    @nm_not_called()
    def test_create_user_salsa_not_dd(self) -> None:
        self.setup_provider("salsa")
        signon = DebianSignon(self.request)
        user = signon.create_user_from_identity(self.identity)
        assert user is not None
        self.assertQuerySetEqual(user.debusine_groups.all(), [])

    @nm_not_called()
    def test_create_user_salsa_dd(self) -> None:
        self.identity.claims["groups_direct"] = ["debian"]
        self.identity.save()
        self.setup_provider("salsa", add_to_group={"debian": "debian/Debian"})
        signon = DebianSignon(self.request)
        user = signon.create_user_from_identity(self.identity)
        assert user is not None
        signon.add_user_to_groups(
            user, self.identity, self.provider.options["add_to_group"]
        )
        assert user is not None
        self.assertQuerySetEqual(
            user.debusine_groups.all(), [self.debian_group]
        )

    @nm_not_called()
    def test_create_user_gitlab_not_contributor(self) -> None:
        self.setup_provider("gitlab")
        signon = DebianSignon(self.request)
        user = signon.create_user_from_identity(self.identity)
        assert user is not None
        self.assertQuerySetEqual(user.debusine_groups.all(), [])

    @nm_not_called()
    def test_create_user_no_user(self) -> None:
        """Test when Signon.create_user_from_identity fails."""
        with mock.patch(
            "debusine.server.signon.signon.Signon.create_user_from_identity",
            return_value=None,
        ):
            signon = DebianSignon(self.request)
            user = signon.create_user_from_identity(self.identity)
        self.assertIsNone(user)

    @nm_not_called()
    def test_create_user_no_provider(self) -> None:
        """Test when the identity has no provider."""
        with (
            mock.patch(
                "debusine.server.signon.signon"
                ".Signon.get_provider_for_identity",
                return_value=None,
            ),
        ):
            signon = DebianSignon(self.request)
            user = signon.create_user_from_identity(self.identity)
        assert user is not None
        self.assertQuerySetEqual(user.debusine_groups.all(), [])

    @responses.activate
    def test_fetch_nm_claims_user_not_found(self) -> None:
        responses.add(
            responses.GET,
            "https://nm.debian.org/api/salsa_status/123",
            json={},
            status=status.HTTP_404_NOT_FOUND,
        )
        signon = DebianSignon(self.request)
        with self.assertLogs("debusine.server.signon") as log:
            self.assertEqual(signon.fetch_nm_claims(123), {})

        self.assertEqual(
            log.output,
            [
                "INFO:debusine.server.signon:"
                "https://nm.debian.org/api/salsa_status/123:"
                " user is not known in nm.debian.org"
            ],
        )

    @responses.activate
    def test_fetch_nm_claims_missing_status(self) -> None:
        responses.add(
            responses.GET,
            "https://nm.debian.org/api/salsa_status/123",
            json={},
        )
        signon = DebianSignon(self.request)
        with self.assertLogs("debusine.server.signon") as log:
            self.assertEqual(signon.fetch_nm_claims(123), {})

        self.assertEqual(
            log.output,
            [
                "WARNING:debusine.server.signon:"
                "https://nm.debian.org/api/salsa_status/123:"
                " user record has no 'status' information"
            ],
        )

    @responses.activate
    def test_fetch_nm_claims_error(self) -> None:
        responses.add(
            responses.GET,
            "https://nm.debian.org/api/salsa_status/123",
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )
        signon = DebianSignon(self.request)
        with self.assertLogs("debusine.server.signon") as log:
            self.assertEqual(signon.fetch_nm_claims(123), {})

        self.assertEqual(
            log.output,
            [
                "WARNING:debusine.server.signon:"
                "https://nm.debian.org/api/salsa_status/123:"
                " cannot fetch user information: 500 (Internal Server Error)"
            ],
        )

    @responses.activate
    def test_fetch_nm_claims(self) -> None:
        responses.add(
            responses.GET,
            "https://nm.debian.org/api/salsa_status/123",
            json={"status": "dm"},
        )
        signon = DebianSignon(self.request)
        with self.assertNoLogs("debusine.server.signon"):
            self.assertEqual(signon.fetch_nm_claims(123), {"nm_status": "dm"})



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View for Lintian work request."""

import json
import uuid
from collections.abc import Iterable
from typing import Any, TypeAlias, override

from debusine.artifacts import LintianArtifact
from debusine.artifacts.models import BaseArtifactDataModel, TaskTypes
from debusine.db.models import Artifact, FileInArtifact
from debusine.db.tasks import DBTask
from debusine.tasks import Lintian
from debusine.tasks.models import LintianData, LintianDynamicData
from debusine.web.views.work_request import Tab, WorkRequestPlugin

DBLintian: TypeAlias = DBTask[LintianData, LintianDynamicData]


class LintianAnalysisOccurrence(BaseArtifactDataModel):
    """A location where a Lintian tag occurred."""

    pointer: str
    note: str
    comment: str


class LintianAnalysisTag(BaseArtifactDataModel):
    """Information about a Lintian tag, with its occurrences grouped."""

    explanation: str
    severity: str
    uuid: uuid.UUID
    occurrences: list[LintianAnalysisOccurrence]


class LintianTab(Tab):
    """Lintian specialized tab."""

    def __init__(self, task: DBLintian) -> None:
        """Define this tab."""
        super().__init__(name="lintian", label="Lintian")
        self.work_request = task.work_request
        self.task = task

    @classmethod
    def _artifacts_to_tags(
        cls, artifacts: Iterable[Artifact]
    ) -> dict[str, dict[str, LintianAnalysisTag]]:
        """
        Receive a list of artifacts and return dict with file names and tags.

        :param artifacts: each contains an "analysis.json" file with "tags"
          key, with a list of tags. Each tag is such as:
            {
               "severity": "pedantic",
               "package": "hello",
               "tag": "license-problem",
               "explanation": "the explanation",
               "comment": "a comment",
               "note": "bar",
               "pointer": "debian/copyright"
            }
        :return: a dictionary of filenames to dictionary containing the tags
          grouped by occurrence. For example:
            {
                "hello_2.10-3.dsc": {
                    "license-problem": LintianAnalysisTag(
                        explanation="the explanation",
                        severity="pedantic",
                        uuid="used_by_the_view",
                        occurrences=[
                            LintianAnalysisOccurrence(
                                pointer="debian/copyright"
                                note="bar",
                                comment="a comment",
                            ),
                        ]
                    }
                }
            }

        """  # noqa: RST201, RST203, RST301
        # "tags": intermediate data structure: contains each file (.dsc,
        # .deb, etc.) with each tag for the file with all the information
        # (explanation, pointer, comment, etc.).
        tags: dict[str, dict[str, list[LintianAnalysisOccurrence]]] = {}

        # "tags_information": intermediate data structure with each tag_name
        # containing a dictionary with the common data: explanation, pointer
        # and severity
        tags_information: dict[str, dict[str, str]] = {}

        for analysis in cls._get_analyses(artifacts):
            for tag in analysis["tags"]:
                package = tag["package"]
                filename = analysis["summary"]["package_filename"][package]

                tag_name = tag["tag"]
                # Explanation and severity is the same for each instance
                # of this tag (incorporated in tags_with_information).
                tag_occurrence = LintianAnalysisOccurrence(
                    pointer=tag["pointer"],
                    note=tag["note"].replace("\\n", "\n"),
                    comment=tag["comment"],
                )
                tags_information[tag_name] = {
                    "explanation": tag["explanation"],
                    "severity": tag["severity"],
                }

                tags.setdefault(filename, {}).setdefault(tag_name, []).append(
                    tag_occurrence
                )

        # "tags_with_information" is the final structure as explained
        # in the docstring
        tags_with_information: dict[str, dict[str, LintianAnalysisTag]] = {}

        for filename, tags_in_filename in tags.items():
            tags_with_information[filename] = {}
            for tag_name, occurrences in tags_in_filename.items():
                tags_with_information[filename][tag_name] = LintianAnalysisTag(
                    **tags_information[tag_name],
                    uuid=uuid.uuid4(),
                    occurrences=occurrences,
                )

        return tags_with_information

    @classmethod
    def _get_summaries(
        cls, artifacts: Iterable[Artifact]
    ) -> list[dict[str, Any]]:
        """Return list of all the artifacts.data["summary"] in artifacts."""
        return [artifact.data["summary"] for artifact in artifacts]

    @staticmethod
    def _get_analyses(artifacts: Iterable[Artifact]) -> list[dict[str, Any]]:
        """Return a list with all the parsed analysis.json in artifacts."""
        analyses: list[dict[str, Any]] = []

        for artifact in artifacts:
            try:
                analysis_in_artifact = artifact.fileinartifact_set.get(
                    path="analysis.json", complete=True
                )
            except FileInArtifact.DoesNotExist:
                continue

            scope = artifact.workspace.scope
            fileobj = analysis_in_artifact.file
            file_backend = scope.download_file_backend(fileobj)

            with file_backend.get_stream(fileobj) as file:
                analysis = json.loads(file.read())

            analyses.append(analysis)

        return analyses

    @staticmethod
    def _find_lintian_txt_url_path(artifacts: Iterable[Artifact]) -> str | None:
        """
        Return the URL path for the lintian.txt file.

        The path is for any of the lintian.txt found in artifacts.
        """
        for artifact in artifacts:
            for file_in_artifact in artifact.fileinartifact_set.filter(
                complete=True
            ).order_by("id"):
                if file_in_artifact.path == Lintian.CAPTURE_OUTPUT_FILENAME:
                    return (
                        artifact.get_absolute_url_download()
                        + Lintian.CAPTURE_OUTPUT_FILENAME
                    )

        return None

    @classmethod
    def _count_by_severity(
        cls, artifacts: Iterable[Artifact]
    ) -> dict[str, int]:
        tags_count_severity = Lintian.generate_severity_count_zero()
        tags_count_severity["overridden"] = 0

        for summary in cls._get_summaries(artifacts):
            for severity, count in summary.get(
                "tags_count_by_severity", {}
            ).items():
                tags_count_severity[severity] += count

        return tags_count_severity

    @staticmethod
    def _list_files(artifact: Artifact | None) -> list[str]:
        if artifact is not None:
            return [
                file.path
                for file in artifact.fileinartifact_set.filter(
                    complete=True
                ).order_by("path")
            ]
        else:
            return []

    @classmethod
    def _get_request_data(cls, task: DBLintian) -> dict[str, Any]:
        """Prepare the requested data to be displayed."""
        dynamic_data = task.dynamic_data

        request_data: dict[str, Any] = {
            "source_artifact": {
                "lookup": task.data.input.source_artifact,
                "id": None,
                "files": [],
            },
            "binary_artifacts": {
                "lookup": task.data.input.binary_artifacts.export(),
                "artifacts": [],
            },
        }

        if dynamic_data is not None:
            if dynamic_data.input_source_artifact_id is not None:
                try:
                    source_artifact = Artifact.objects.get(
                        id=dynamic_data.input_source_artifact_id
                    )
                except Artifact.DoesNotExist:
                    pass
                else:
                    request_data["source_artifact"].update(
                        {
                            "id": source_artifact.id,
                            "files": cls._list_files(source_artifact),
                            "artifact": source_artifact,
                        }
                    )

            binary_artifacts = {
                artifact.id: artifact
                for artifact in Artifact.objects.filter(
                    id__in=dynamic_data.input_binary_artifacts_ids
                )
            }
            for binary_artifact_id in dynamic_data.input_binary_artifacts_ids:
                binary_artifact = binary_artifacts.get(binary_artifact_id)
                request_data["binary_artifacts"]["artifacts"].append(
                    {
                        "id": binary_artifact_id,
                        "files": cls._list_files(binary_artifact),
                        "artifact": binary_artifact,
                    }
                )

        return request_data

    @override
    def get_context_data(self) -> dict[str, Any]:
        task = self.task

        artifacts = self.work_request.artifact_set.filter(
            category=LintianArtifact._category
        ).order_by("id")

        lintian_txt_path = self._find_lintian_txt_url_path(artifacts)

        return {
            "request_data": self._get_request_data(task),
            "result": self.work_request.result,
            "lintian_txt_path": lintian_txt_path,
            "fail_on_severity": task.data.fail_on_severity,
            "tags": self._artifacts_to_tags(artifacts),
            "tags_count_severity": self._count_by_severity(artifacts),
        }


class LintianViewWorkRequestPlugin(WorkRequestPlugin):
    """View for Lintian work request."""

    task_type = TaskTypes.WORKER
    task_name = "lintian"
    task: DBLintian

    @override
    def add_tabs(self, tabs: list[Tab]) -> None:
        tabs.insert(0, LintianTab(self.task))

    @override
    def do_get_description_data(self) -> dict[str, Any]:
        """Parse metadata."""
        dynamic_data = self.task.dynamic_data
        assert dynamic_data is not None

        data = self.task.data

        return {
            "package_name": dynamic_data.subject,
            "source_artifact_id": dynamic_data.input_source_artifact_id,
            "binary_artifacts_ids": dynamic_data.input_binary_artifacts_ids,
            "environment_id": dynamic_data.environment_id,
            "fail_on_severity": data.fail_on_severity,
            "build_architecture": data.build_architecture,
            "target_distribution": data.target_distribution,
            "include_tags": data.include_tags,
            "exclude_tags": data.exclude_tags,
        }

    @override
    def get_context_data(self) -> dict[str, Any]:
        return {
            "description_template": "web/_lintian-description.html",
            "description_data": self.get_description_data(),
        }



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db workers."""

import copy
import datetime as dt
import hashlib
from typing import (
    Any,
    Optional,
    TYPE_CHECKING,
    TypedDict,
    cast,
    overload,
    override,
)

import pgtrigger
from django.db import IntegrityError, models, transaction
from django.db.models import (
    CheckConstraint,
    Count,
    Exists,
    F,
    JSONField,
    Max,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
)
from django.db.models.expressions import Case, When
from django.db.models.functions import Extract, Greatest, Now
from django.urls import reverse
from django.utils import timezone
from django.utils.text import slugify

from debusine.db.models import WorkRequest
from debusine.db.models.auth import Token
from debusine.db.models.worker_pools import WorkerPool
from debusine.tasks.models import WorkerType

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext import WithAnnotations
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.web.views.ui.workers import WorkerUI
else:
    TypedModelMeta = object


class IdleTimeDict(TypedDict):
    """Additional fields returned by WorkerQuerySet.with_idle_time."""

    idle_time: float


class WorkerQuerySet[A](QuerySet["Worker", A]):
    """Custom QuerySet for Worker."""

    def connected(self) -> "WorkerQuerySet[Any]":
        """Return connected workers."""
        # WorkerConsumer.REQUEST_DYNAMIC_METADATA_SECONDS = 3600, so this
        # seems like a reasonable threshold beyond which to assume that a
        # worker is no longer really connected even if it hasn't explicitly
        # disconnected.
        threshold = Now() - dt.timedelta(hours=1, minutes=5)
        return (
            self.filter(connected_at__isnull=False)
            .filter(
                Q(worker_type=WorkerType.CELERY)
                | Q(dynamic_metadata_updated_at__gte=threshold)
            )
            .order_by("-dynamic_metadata_updated_at")
        )

    def active(self) -> "WorkerQuerySet[Any]":
        """Exclude inactive workers from worker pools."""
        return self.exclude(
            worker_pool__isnull=False, instance_created_at__isnull=True
        )

    def waiting_for_work_request(self) -> "WorkerQuerySet[Any]":
        """
        Return workers that can be assigned a new work request.

        The workers with fewer associated pending or running work requests
        than their concurrency level could take more work right now and are
        thus waiting for a work request.

        Worker's token must be enabled.
        """
        # Import here to prevent circular imports
        from debusine.db.models.work_requests import WorkRequest

        running_work_request_count = Count(
            'assigned_work_requests',
            filter=Q(
                assigned_work_requests__status__in=[
                    WorkRequest.Statuses.RUNNING,
                    WorkRequest.Statuses.PENDING,
                ]
            ),
        )
        workers = (
            self.connected()
            .order_by(F('worker_pool').asc(nulls_first=True), 'connected_at')
            .annotate(count_running=running_work_request_count)
            .filter(count_running__lt=F("concurrency"))
            .filter(Q(worker_type=WorkerType.CELERY) | Q(token__enabled=True))
        )

        return workers

    def with_idle_time(
        self,
    ) -> "WorkerQuerySet[WithAnnotations[Worker, IdleTimeDict]]":
        """
        Annotate workers with `idle_time`.

        `Worker.idle_time` will be 0 for workers currently assigned a work
        request, or the number of seconds the worker has been spending without
        a work request assigned.
        """
        tasks = WorkRequest.objects.filter(worker=OuterRef("pk"))

        active_tasks = tasks.filter(
            status__in=(
                WorkRequest.Statuses.RUNNING,
                WorkRequest.Statuses.PENDING,
            )
        )

        last_tasks = tasks.filter(
            completed_at__gt=F("worker__instance_created_at")
        )

        last_completed_at = Subquery(
            last_tasks.annotate(last_completed_at=Max("completed_at")).values(
                "last_completed_at"
            )
        )
        # Work-around https://code.djangoproject.com/ticket/28296
        last_completed_at.query.set_group_by()
        last_completed_at.query.clear_ordering()

        return self.annotate(
            idle_time=Case(
                When(
                    Exists(active_tasks),
                    then=0,
                ),
                When(
                    Exists(last_tasks),
                    then=Extract(Now() - last_completed_at, "epoch"),
                ),
                default=Extract(
                    Now() - Greatest("registered_at", "instance_created_at"),
                    "epoch",
                ),
            )
        )

    def get_worker_by_token_key_or_none(
        self, token_key: str
    ) -> Optional["Worker"]:
        """Return a Worker identified by its associated secret token."""
        try:
            token_hash = hashlib.sha256(token_key.encode()).hexdigest()
            return cast("Worker", self.get(token__hash=token_hash))
        except Worker.DoesNotExist:
            return None

    def get_worker_or_none(self, worker_name: str) -> Optional["Worker"]:
        """Return the worker with worker_name or None."""
        try:
            return cast("Worker", self.get(name=worker_name))
        except Worker.DoesNotExist:
            return None


class WorkerManager(models.Manager["Worker"]):
    """Manager for Worker model."""

    @staticmethod
    def _generate_unique_name(name: str, counter: int) -> str:
        """Return name slugified adding "-counter" if counter != 1."""
        new_name = slugify(name.replace('.', '-'))

        if counter != 1:
            new_name += f'-{counter}'

        return new_name

    @overload
    def create_with_fqdn(
        self,
        fqdn: str,
        *,
        token: Token,
        activation_token: None = None,
        worker_type: WorkerType = WorkerType.EXTERNAL,
        worker_pool: WorkerPool | None = None,
    ) -> "Worker": ...

    @overload
    def create_with_fqdn(
        self,
        fqdn: str,
        *,
        token: None = None,
        activation_token: Token,
        worker_type: WorkerType = WorkerType.EXTERNAL,
        worker_pool: WorkerPool | None = None,
    ) -> "Worker": ...

    def create_with_fqdn(
        self,
        fqdn: str,
        *,
        token: Token | None = None,
        activation_token: Token | None = None,
        worker_type: WorkerType = WorkerType.EXTERNAL,
        worker_pool: WorkerPool | None = None,
    ) -> "Worker":
        """Return a new Worker with its name based on fqdn, with token."""
        counter = 1

        while True:
            name = self._generate_unique_name(fqdn, counter)
            try:
                with transaction.atomic():
                    return self.create(
                        name=name,
                        token=token,
                        activation_token=activation_token,
                        worker_type=worker_type,
                        registered_at=timezone.now(),
                        worker_pool=worker_pool,
                    )
            except IntegrityError as e:
                if "db_worker_name_key" not in str(e):
                    raise
                counter += 1

    def create_pool_members(
        self,
        worker_pool: WorkerPool,
        count: int,
    ) -> None:
        """
        Create count new External Workers in worker_pool.

        Use throw-away activation tokens; when workers start, they'll
        exchange their activation tokens for full worker tokens.
        """
        created = 0
        index = 1
        while created < count:
            with transaction.atomic():
                try:
                    self.create(
                        name=f"{worker_pool.name}-{index:03}",
                        activation_token=(
                            Token.objects.create_worker_activation()
                        ),
                        worker_type=WorkerType.EXTERNAL,
                        worker_pool=worker_pool,
                        registered_at=timezone.now(),
                    )
                    created += 1
                except IntegrityError:
                    pass
            index += 1

    def get_or_create_scheduler(self) -> "Worker":
        """
        Return a new Worker representing the scheduler.

        The scheduler of course can't dispatch tasks to itself, but
        representing it as a worker is convenient because it allows the
        scheduler to call back to the server to emit metrics.
        """
        try:
            return self.get(name="scheduler", worker_type=WorkerType.SCHEDULER)
        except Worker.DoesNotExist:
            return self.create(
                name="scheduler",
                worker_type=WorkerType.SCHEDULER,
                registered_at=timezone.now(),
            )

    def get_or_create_celery(self) -> "Worker":
        """Return a new Worker representing the Celery task queue."""
        try:
            return self.get(name="celery", worker_type=WorkerType.CELERY)
        except Worker.DoesNotExist:
            return self.create(
                name="celery",
                worker_type=WorkerType.CELERY,
                registered_at=timezone.now(),
            )

    def mark_all_disconnected(self) -> None:
        """
        Mark all non-Celery workers as disconnected.

        When debusine-server starts, depending on how it exited, workers
        might still be registered as connected in the database.  To avoid
        confusion, mark them as disconnected until they explicitly connect
        again.
        """
        Worker.objects.connected().exclude(
            worker_type=WorkerType.CELERY
        ).update(connected_at=None)


class Worker(models.Model):
    """Database model of a worker."""

    name = models.SlugField(
        unique=True,
        help_text='Human readable name of the worker based on the FQDN',
    )
    registered_at = models.DateTimeField()
    connected_at = models.DateTimeField(blank=True, null=True)
    instance_created_at = models.DateTimeField(blank=True, null=True)

    # This is the token used by the Worker to authenticate
    # Users have their own tokens - this is specific to a single worker.
    token = models.OneToOneField(
        Token, null=True, on_delete=models.PROTECT, related_name="worker"
    )
    # This is used to grant workers the ability to bootstrap authentication
    # for themselves, in situations where their provisioning data may be
    # revealed to untrusted code later.  The activation token will be
    # deleted once it has been used.
    activation_token = models.OneToOneField(
        Token,
        null=True,
        on_delete=models.PROTECT,
        related_name="activating_worker",
    )
    # Set by debusine-admin worker edit_metadata.
    # Note: contents will be displayed to users in the web UI
    static_metadata = JSONField(default=dict, blank=True)
    # Information about features supported by workers.
    # Note: contents will be displayed to users in the web UI
    dynamic_metadata = JSONField(default=dict, blank=True)
    # Information needed to track cloud workers.
    # Note: contents will be displayed to users in the web UI
    dynamic_metadata_updated_at = models.DateTimeField(blank=True, null=True)

    worker_type = models.CharField(
        choices=WorkerType.choices, default=WorkerType.EXTERNAL, editable=False
    )
    # Only Celery workers currently support concurrency levels greater than
    # 1.
    concurrency = models.PositiveIntegerField(
        default=1,
        help_text="Number of tasks this worker can run simultaneously",
    )

    worker_pool = models.ForeignKey(
        WorkerPool, blank=True, null=True, on_delete=models.PROTECT
    )
    worker_pool_data = JSONField(null=True, blank=True)

    objects = WorkerManager.from_queryset(WorkerQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        triggers = [
            # Both Worker and Token need check constraints on the foreign
            # keys between them, so keep copies in sync using triggers.
            *(
                pgtrigger.Trigger(
                    name=f"db_worker_sync_unset_{worker_column}",
                    operation=(
                        pgtrigger.UpdateOf(f"{worker_column}_id")  # type: ignore[no-untyped-call]
                        | pgtrigger.Delete
                    ),
                    when=pgtrigger.Before,
                    condition=pgtrigger.Q(
                        **{f"old__{worker_column}__isnull": False}
                    ),
                    func=" ".join(
                        f"""
                        UPDATE db_token
                            SET {token_column}_id = NULL
                            WHERE db_token.id = OLD.{worker_column}_id;
                        IF TG_OP = 'DELETE' THEN
                            RETURN OLD;
                        ELSE
                            RETURN NEW;
                        END IF;
                        """.split()
                    ),
                )
                for worker_column, token_column in (
                    ("token", "worker"),
                    ("activation_token", "activating_worker"),
                )
            ),
            *(
                pgtrigger.Trigger(
                    name=f"db_worker_sync_set_{worker_column}",
                    operation=(
                        pgtrigger.Insert
                        | pgtrigger.UpdateOf(f"{worker_column}_id")  # type: ignore[no-untyped-call]
                    ),
                    when=pgtrigger.After,
                    condition=pgtrigger.Q(
                        **{f"new__{worker_column}__isnull": False}
                    ),
                    func=" ".join(
                        f"""
                        UPDATE db_token
                            SET {token_column}_id = NEW.id
                            WHERE db_token.id = NEW.{worker_column}_id;
                        RETURN NEW;
                        """.split()
                    ),
                )
                for worker_column, token_column in (
                    ("token", "worker"),
                    ("activation_token", "activating_worker"),
                )
            ),
        ]
        constraints = [
            # Non-Celery workers must have a token.
            CheckConstraint(
                name="%(app_label)s_%(class)s_celery_or_token",
                check=Q(
                    worker_type__in={WorkerType.CELERY, WorkerType.SCHEDULER}
                )
                | Q(activation_token__isnull=False)
                | Q(token__isnull=False)
                # Exclude worker_pool tombstones
                | Q(
                    worker_pool_id__isnull=False,
                    instance_created_at__isnull=True,
                ),
            )
        ]

    @override
    def __str__(self) -> str:
        """Return the id and name of the Worker."""
        return f"Id: {self.id} Name: {self.name}"

    def get_absolute_url(self) -> str:
        """Return an absolute URL to view this worker."""
        return reverse("workers:detail", kwargs={"name": self.name})

    def mark_disconnected(self) -> None:
        """Update and save relevant Worker fields after disconnecting."""
        self.connected_at = None
        self.save()

    def running_work_requests(self) -> QuerySet["WorkRequest"]:
        """Return queryset of work requests running on this worker."""
        return self.assigned_work_requests.filter(
            status=WorkRequest.Statuses.RUNNING
        ).order_by("id")

    def mark_connected(self) -> None:
        """Update and save relevant Worker fields after connecting."""
        from debusine.server.scheduler import schedule_soon

        self.connected_at = timezone.now()
        self.save()
        schedule_soon()

    def connected(self) -> bool:
        """Return True if the Worker is connected."""
        return self.connected_at is not None

    def is_busy(self) -> bool:
        """Return True if the Worker has any active work requests."""
        # Import here to prevent circular imports
        from debusine.db.models.work_requests import WorkRequest

        return (
            WorkRequest.objects.running(worker=self)
            | WorkRequest.objects.pending(worker=self)
        ).exists()

    def is_at_capacity(self) -> bool:
        """
        Return True if the Worker is busy with work requests.

        A Worker is at capacity if it has as many running or pending work
        requests as its concurrency level.
        """
        # Import here to prevent circular imports
        from debusine.db.models.work_requests import WorkRequest

        return (
            WorkRequest.objects.running(worker=self)
            | WorkRequest.objects.pending(worker=self)
        ).count() >= self.concurrency

    def metadata(self) -> dict[str, Any]:
        """
        Return all metadata with static_metadata and dynamic_metadata merged.

        If the same key is in static_metadata and dynamic_metadata:
        static_metadata takes priority.
        The provided_tags key is merged using get_provided_worker_tags.
        """
        from debusine.server.scheduler import get_provided_worker_tags

        metadata = {
            **copy.deepcopy(self.dynamic_metadata),
            **copy.deepcopy(self.static_metadata),
        }

        metadata["provided_tags"] = sorted(get_provided_worker_tags(self))

        return metadata

    def set_dynamic_metadata(self, metadata: dict[str, Any]) -> None:
        """Save metadata and update dynamic_metadata_updated_at."""
        self.dynamic_metadata = metadata
        self.dynamic_metadata_updated_at = timezone.now()
        self.save()

    def ui(self, request: "HttpRequest") -> "WorkerUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.workers import WorkerUI

        return WorkerUI.get(request, self)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs related to collections in workspaces."""

from django.urls import include, path

import debusine.web.views.collections as views

app_name = "collections"

urlpatterns = [
    path(
        "",
        views.CollectionListView.as_view(),
        name="list",
    ),
    path(
        "create/",
        views.CollectionCreateView.as_view(),
        name="create",
    ),
    path(
        "<str:ccat>/",
        views.CollectionCategoryListView.as_view(),
        name="category_list",
    ),
    path(
        "<str:ccat>/<str:cname>/",
        views.CollectionDetailView.as_view(),
        name="detail",
    ),
    path(
        "<str:ccat>/<str:cname>/update/",
        views.CollectionUpdateView.as_view(),
        name="update",
    ),
    path(
        "<str:ccat>/<str:cname>/search/",
        views.CollectionSearchView.as_view(),
        name="search",
    ),
    path(
        "<str:ccat>/<str:cname>/delete/",
        views.CollectionDeleteView.as_view(),
        name="delete",
    ),
    path(
        "<str:ccat>/<str:cname>/item/<str:iid>/<path:iname>/",
        views.CollectionItemDetailView.as_view(),
        name="item_detail",
    ),
    path(
        "<str:ccat>/<str:cname>/relation/",
        include(
            "debusine.web.urls.collection_relations",
        ),
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the WorkflowInternalManager."""

from typing import override
from unittest import mock

from django.db import IntegrityError

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
)
from debusine.db.models import Collection, CollectionItem
from debusine.server.collections import (
    CollectionManagerInterface,
    ItemAdditionError,
    WorkflowInternalManager,
)
from debusine.test.django import TestCase


class WorkflowInternalManagerTests(TestCase):
    """Tests for WorkflowInternalManager."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.user = self.playground.get_default_user()

        self.workspace = self.playground.get_default_workspace()

        self.collection = Collection.objects.create(
            name="Test",
            category=CollectionCategory.WORKFLOW_INTERNAL,
            workspace=self.workspace,
        )

        self.manager = WorkflowInternalManager(collection=self.collection)

    def test_do_add_bare_data_no_name(self) -> None:
        """`do_add_bare_data` requires an item name."""
        with self.assertRaisesRegex(
            ItemAdditionError,
            "Adding to debusine:workflow-internal requires an item name",
        ):
            self.manager.add_bare_data(BareDataCategory.TEST, user=self.user)

    def test_do_add_bare_data_raise_item_addition_error(self) -> None:
        """Test do_add_bare_data raise error on Integrity issue."""
        with mock.patch(
            "debusine.db.models.CollectionItem.objects.create_from_bare_data"
        ) as mocked:
            mocked.side_effect = IntegrityError
            with self.assertRaises(ItemAdditionError):
                self.manager.add_bare_data(
                    BareDataCategory.TEST, user=self.user, name="actiontest"
                )

    def test_do_add_bare_data_replace(self) -> None:
        """`do_add_bare_data` can replace an existing artifact."""
        collection_item = self.manager.add_bare_data(
            BareDataCategory.TEST, user=self.user, name="test"
        )

        collection_item2 = self.manager.add_bare_data(
            BareDataCategory.TEST,
            user=self.user,
            data={"foo": "bar"},
            name="test",
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.name, "test")
        self.assertEqual(collection_item.child_type, CollectionItem.Types.BARE)
        self.assertEqual(collection_item.data, {})
        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertIsNotNone(collection_item.removed_at)
        self.assertEqual(collection_item2.name, "test")
        self.assertEqual(collection_item2.child_type, CollectionItem.Types.BARE)
        self.assertEqual(collection_item2.data, {"foo": "bar"})
        self.assertIsNone(collection_item2.removed_at)

    def test_do_add_bare_data_invalid_field_name(self) -> None:
        r"""Only debusine:promise can have data keys starting with ^promise_."""
        expected_message = (
            'Fields starting with "promise_" are not allowed'
            ' for category "debusine:test"'
        )
        with self.assertRaisesRegex(ItemAdditionError, expected_message):
            self.manager.add_bare_data(
                BareDataCategory.TEST,
                user=self.user,
                name="test",
                data={"promise_id": 100},
            )

    def test_do_add_bare_data_replace_nonexistent(self) -> None:
        """Replacing a nonexistent bare data item is allowed."""
        collection_item = self.manager.add_bare_data(
            BareDataCategory.TEST, user=self.user, name="test", replace=True
        )

        self.assertEqual(collection_item.name, "test")
        self.assertEqual(collection_item.child_type, CollectionItem.Types.BARE)

    def test_do_add_artifact_no_name(self) -> None:
        """`do_add_artifact` requires an item name."""
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            "Adding to debusine:workflow-internal requires an item name",
        ):
            self.manager.add_artifact(artifact_1, user=self.user)

    def test_do_add_artifact_raise_item_addition_error(self) -> None:
        """Test do_add_artifact raise error on Integrity issue."""
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )
        with mock.patch(
            "debusine.db.models.CollectionItem.objects.create_from_artifact"
        ) as mocked:
            mocked.side_effect = IntegrityError
            with self.assertRaises(ItemAdditionError):
                self.manager.add_artifact(
                    artifact_1, user=self.user, name="actiontest"
                )

    def test_do_add_artifact_replace(self) -> None:
        """`do_add_artifact` can replace an existing artifact."""
        workflow = self.playground.create_workflow()

        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )
        collection_item = self.manager.add_artifact(
            artifact_1, user=self.user, name="test"
        )
        artifact_2, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )

        collection_item2 = self.manager.add_artifact(
            artifact_2,
            user=self.user,
            workflow=workflow,
            name="test",
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, artifact_1)
        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertEqual(collection_item.removed_by_workflow, workflow)
        self.assertIsNotNone(collection_item.removed_at)

        self.assertEqual(collection_item2.name, "test")
        self.assertEqual(collection_item2.artifact, artifact_2)
        self.assertIsNone(collection_item2.removed_at)

    def test_do_add_artifact_with_promise_variable_raises_error(self) -> None:
        r"""
        Test adding artifact raises ItemAdditionError.

        The artifact's variables cannot contain keys starting with promise\\_.
        """
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )

        # Variables contain an invalid key starting with 'promise_'
        variables = {
            "promise_invalid_key": "some_value",
            "valid_key": "another_value",
        }

        with self.assertRaisesRegex(
            ItemAdditionError,
            'Keys starting with "promise_" are not allowed in variables',
        ):
            self.manager.add_artifact(
                artifact_1, user=self.user, variables=variables, name="test"
            )

    def test_do_add_artifact_with_variables_copied(self) -> None:
        """Test adding artifact copies the variables into CollectionItem."""
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )

        # Variables contain an invalid key starting with 'promise_'
        variables = {
            "valid_key_1": "value_1",
            "valid_key_2": "value_2",
        }

        collection_item = self.manager.add_artifact(
            artifact_1, user=self.user, variables=variables, name="test"
        )

        self.assertEqual(collection_item.data, variables)

    def test_do_add_artifact_replace_nonexistent(self) -> None:
        """Replacing a nonexistent artifact is allowed."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )

        collection_item = self.manager.add_artifact(
            artifact, user=self.user, name="test", replace=True
        )

        self.assertEqual(collection_item.name, "test")
        self.assertEqual(collection_item.artifact, artifact)

    def test_do_add_artifact_replace_bare_data(self) -> None:
        """Replacing bare data with an artifact is allowed."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST, data={}
        )
        self.manager.add_bare_data(
            category=BareDataCategory.TEST, user=self.user, name="test", data={}
        )

        self.manager.add_artifact(
            artifact, user=self.user, name="test", replace=True
        )

    def test_lookup_unexpected_format_raise_lookup_error(self) -> None:
        """Test lookup raise LookupError: invalid format."""
        msg = '^Unexpected lookup format: "foo:bar"$'

        with self.assertRaisesRegex(LookupError, msg):
            self.manager.lookup("foo:bar")

    def test_equal(self) -> None:
        """Test __eq__."""
        collection = Collection.objects.create(
            name="TestDebianSuite",
            category=CollectionCategory.SUITE,
            workspace=self.workspace,
        )
        manager_debian_suite = CollectionManagerInterface.get_manager_for(
            collection
        )
        manager_debian_suite2 = CollectionManagerInterface.get_manager_for(
            collection
        )
        self.assertEqual(manager_debian_suite, manager_debian_suite2)

        collection = Collection.objects.create(
            name="TestDebianQAResults",
            category=CollectionCategory.QA_RESULTS,
            workspace=self.workspace,
        )
        manager_debian_qa_results = CollectionManagerInterface.get_manager_for(
            collection
        )
        self.assertNotEqual(manager_debian_suite, manager_debian_qa_results)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for view_utils."""

import json

import yaml
from django.template.defaultfilters import filesizeformat

from debusine.db.models import FileInArtifact, FileStore
from debusine.test.django import TestCase
from debusine.web.views import view_utils


class ReadFileTests(TestCase):
    """Tests for read_file() function."""

    def create_file_in_backend(
        self,
        *,
        contents: bytes = b"sample-data",
        path: str = "/usr/bin/test",
        complete: bool = True,
    ) -> FileInArtifact:
        """
        Create a FileInArtifact and add it into the backend.

        This is intended for read_file() tests, where the file must exist in
        the backend.
        """
        artifact, _ = self.playground.create_artifact()
        file_backend = FileStore.default().get_backend_object()

        file = self.playground.create_file_in_backend(
            backend=file_backend,
            contents=contents,
        )

        file_in_artifact = FileInArtifact(
            file=file,
            artifact=artifact,
            path=path,
            complete=complete,
        )

        return file_in_artifact

    def test_read_file(self) -> None:
        file_in_artifact = self.create_file_in_backend(contents=b"The contents")

        self.assertEqual(
            view_utils.read_file(file_in_artifact, max_file_size=1024),
            b"The contents",
        )

    def test_read_file_too_big_error(self) -> None:
        max_file_size = 10
        file_in_artifact = self.create_file_in_backend(
            contents=b"a" * (max_file_size + 1), complete=True
        )

        with self.assertRaises(view_utils.TooBigFileInArtifactError) as error:
            view_utils.read_file(file_in_artifact, max_file_size=max_file_size)

        self.assertEqual(
            error.exception.to_html(),
            f'File too big to display '
            f'(size: {filesizeformat(file_in_artifact.file.size)} '
            f'maximum size: {filesizeformat(max_file_size)}). '
            f'<a href="{file_in_artifact.get_absolute_url_download()}">Download'
            '</a> or '
            f'<a href="{file_in_artifact.get_absolute_url_raw()}">view</a> '
            'it raw.',
        )

    def test_read_file_incomplete_error(self) -> None:
        file_in_artifact = self.create_file_in_backend(complete=False)

        with self.assertRaises(
            view_utils.IncompleteFileInArtifactError
        ) as error:
            view_utils.read_file(file_in_artifact, max_file_size=1024)

        self.assertEqual(
            error.exception.to_html(),
            'File "/usr/bin/test" has not completed uploading.',
        )


class TestViewUtils(TestCase):
    """Tests for view_utils functions."""

    CONTENTS_XPATH = 'string(//div[@class="formatted-content"]/pre)'

    def test_as_yaml_round_trip(self) -> None:
        for value in (
            {"package": "hello", "version": "1.0.0"},
            "hello",
            4242,
            42.42,
            True,
            False,
            None,
            [1, "a", 42.42],
        ):
            with self.subTest(value=value, type=type(value).__name__):
                html = view_utils.highlight_as_yaml(value)

                tree = self.assertHTMLValid(html)
                output = tree.xpath(self.CONTENTS_XPATH)

                self.assertEqual(yaml.safe_load(output), value)

    def test_as_json(self) -> None:
        value = {"package": "hello"}

        html = view_utils.highlight_as_json(value)
        tree = self.assertHTMLValid(html)

        output = tree.xpath(self.CONTENTS_XPATH)
        self.assertEqual(json.loads(output), value)

    def test_syntax_highlight(self) -> None:
        value = "some text"

        html = view_utils.syntax_highlight("text", value)
        tree = self.assertHTMLValid(html)

        output = self.get_node_text_normalized(tree)
        self.assertEqual(output, value)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Set common HTTP headers."""

from django.http import HttpRequest, HttpResponseBase
from django.utils.cache import patch_vary_headers
from django.utils.deprecation import MiddlewareMixin


class HeadersMiddleware(MiddlewareMixin):
    """Set common HTTP headers."""

    def process_response(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        response: HttpResponseBase,
    ) -> HttpResponseBase:
        """Process response in middleware."""
        # Don't allow caches to use responses to satisfy requests that have
        # different authentication.
        patch_vary_headers(response, ("Cookie", "Token"))
        return response



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View for Autopkgtest work request."""

import itertools
import re
from enum import StrEnum
from typing import Any, Optional, TypeAlias, assert_never, override

import pydantic
from more_itertools import peekable

from debusine.artifacts import AutopkgtestArtifact
from debusine.artifacts.models import ArtifactCategory, TaskTypes
from debusine.db.models import Artifact, FileInArtifact
from debusine.db.tasks import DBTask
from debusine.tasks.models import AutopkgtestData, AutopkgtestDynamicData
from debusine.web.views import view_utils
from debusine.web.views.artifacts import ArtifactPlugin
from debusine.web.views.files import FileWidget, LogFileWidget
from debusine.web.views.work_request import Tab, WorkRequestPlugin

DBAutopkgtest: TypeAlias = DBTask[AutopkgtestData, AutopkgtestDynamicData]


class AutopkgtestTab(Tab):
    """Autopkgtest specialized tab."""

    def __init__(self, task: DBAutopkgtest) -> None:
        """Define this tab."""
        super().__init__(name="autopkgtest", label="Autopkgtest")
        self.work_request = task.work_request
        self.task = task

    @staticmethod
    def _get_fail_on_scenarios(fail_on: dict[str, bool]) -> list[str]:
        """Process fail_on task data into a list of strings for viewing."""
        return [
            key.replace("_test", "")
            for key, value in fail_on.items()
            if key.endswith("_test") and value
        ]

    @staticmethod
    def _list_files(artifact: Artifact | None) -> list[str]:
        if artifact is not None:
            return [
                file.path
                for file in artifact.fileinartifact_set.filter(
                    complete=True
                ).order_by("path")
            ]
        else:
            return []

    @property
    def _source_artifact(self) -> dict[str, Any]:
        task_data = self.task.data
        source_artifact: dict[str, Any] = {
            "lookup": task_data.input.source_artifact,
            "id": None,
            "artifact": None,
            "files": [],
        }

        if (dynamic_data := self.task.dynamic_data) is not None:
            pk = dynamic_data.input_source_artifact_id
            source_artifact["id"] = pk
            try:
                source_artifact["artifact"] = artifact = Artifact.objects.get(
                    id=pk
                )
            except Artifact.DoesNotExist:
                pass
            else:
                source_artifact["files"] = self._list_files(artifact)
        return source_artifact

    def _get_request_data(self) -> dict[str, Any]:
        """Return data about the request."""
        task = self.task
        task_data = task.data
        dynamic_data = task.dynamic_data

        # TODO: It would be useful to show the request data after default
        # values have been set, but unfortunately these are only set by the
        # task after it loads data from the database and aren't saved
        # anywhere else.
        request_data: dict[str, Any] = {
            "source_artifact": self._source_artifact,
            "source_name": None,
            "source_version": None,
            "binary_artifacts": {
                "lookup": task_data.input.binary_artifacts.export(),
                "artifacts": [],
            },
            "context_artifacts": {
                "lookup": task_data.input.context_artifacts.export(),
                "artifacts": [],
            },
            "build_architecture": task_data.build_architecture,
            "vendor": None,
            "codename": None,
            "backend": task_data.backend,
            "include_tests": task_data.include_tests,
            "exclude_tests": task_data.exclude_tests,
            "debug_level": task_data.debug_level,
            "extra_repositories": task_data.extra_repositories,
            "use_packages_from_base_repository": (
                task_data.use_packages_from_base_repository
            ),
            "extra_environment": task_data.extra_environment,
            "needs_internet": task_data.needs_internet,
            "fail_on_scenarios": self._get_fail_on_scenarios(
                task_data.fail_on.model_dump()
            ),
            "timeout": task_data.timeout,
        }

        # This information is in the result artifact's data if it exists,
        # but it won't necessarily exist - for example, a testbed failure
        # will leave us without a result artifact.
        if dynamic_data is not None:
            try:
                environment_artifact = Artifact.objects.get(
                    id=dynamic_data.environment_id
                )
            except Artifact.DoesNotExist:
                pass
            else:
                if environment_artifact.category in {
                    ArtifactCategory.SYSTEM_TARBALL,
                    ArtifactCategory.SYSTEM_IMAGE,
                }:
                    request_data["vendor"] = environment_artifact.data.get(
                        "vendor"
                    )
                    request_data["codename"] = environment_artifact.data.get(
                        "codename"
                    )

            if source_artifact := request_data["source_artifact"]["artifact"]:
                if source_artifact.category == ArtifactCategory.SOURCE_PACKAGE:
                    request_data.update(
                        {
                            "source_name": source_artifact.data.get("name"),
                            "source_version": source_artifact.data.get(
                                "version"
                            ),
                        }
                    )

            binary_artifacts = {
                artifact.id: artifact
                for artifact in Artifact.objects.filter(
                    id__in=dynamic_data.input_binary_artifacts_ids
                )
            }
            for binary_artifact_id in dynamic_data.input_binary_artifacts_ids:
                binary_artifact = binary_artifacts.get(binary_artifact_id)
                request_data["binary_artifacts"]["artifacts"].append(
                    {
                        "id": binary_artifact_id,
                        "files": self._list_files(binary_artifact),
                        "artifact": binary_artifact,
                    }
                )

            context_artifacts = {
                artifact.id: artifact
                for artifact in Artifact.objects.filter(
                    id__in=dynamic_data.input_context_artifacts_ids
                )
            }
            for context_artifact_id in dynamic_data.input_context_artifacts_ids:
                context_artifact = context_artifacts.get(context_artifact_id)
                request_data["context_artifacts"]["artifacts"].append(
                    {
                        "id": context_artifact_id,
                        "files": self._list_files(context_artifact),
                    }
                )

        request_data["task_description"] = (
            f"{request_data['source_name'] or 'UNKNOWN'}_"
            f"{request_data['source_version'] or 'UNKNOWN'}_"
            f"{request_data['build_architecture']}"
        )
        if request_data["vendor"] and request_data["codename"]:
            request_data["task_description"] += (
                f" in {request_data['vendor']}:{request_data['codename']}"
            )

        return request_data

    @staticmethod
    def _get_result_data(
        artifact: Artifact,
    ) -> list[tuple[str, str, str | None]]:
        """Return a list of tuples describing individual test results."""
        result_data = []
        for name, result in sorted(artifact.data.get("results", {}).items()):
            result_data.append((name, result["status"], result.get("details")))
        return result_data

    @override
    def get_context_data(self) -> dict[str, Any]:
        result_artifact = (
            self.work_request.artifact_set.filter(
                category=AutopkgtestArtifact._category
            )
            .order_by("id")
            .first()
        )

        context_data: dict[str, Any] = {
            "request_data": self._get_request_data(),
            "result": self.work_request.result,
        }
        if result_artifact:
            context_data["result_artifact"] = result_artifact
            context_data["result_data"] = self._get_result_data(result_artifact)

        return context_data


class AutopkgtestViewWorkRequestPlugin(WorkRequestPlugin):
    """View for Autopkgtest work request."""

    task_type = TaskTypes.WORKER
    task_name = "autopkgtest"
    task: DBAutopkgtest

    @override
    def add_tabs(self, tabs: list[Tab]) -> None:
        tabs.insert(0, AutopkgtestTab(self.task))

    @override
    def get_context_data(self) -> dict[str, Any]:
        return {}


class AutopkgtestLogParser:
    """Parse Autopkgtest logs."""

    class Result(StrEnum):
        PASS = "PASS"
        FAIL = "FAIL"
        SKIP = "SKIP"

    class Section(pydantic.BaseModel):
        """Represents a logical section."""

        text: str
        color: str | None = (
            None  # bootstrap color: "primary", "secondary", "success"...
        )
        subsections: list["AutopkgtestLogParser.SubSection"] = []

    class SubSection(Section):
        """Represents a logical subsection with the test result."""

        line: int  # line number, 1-indexed
        result: Optional["AutopkgtestLogParser.Result"] = None
        expanded: bool = False  # subsection's contents is expanded or not
        include_in_table_of_contents: bool = True
        tooltip: str | None = None

    type TestNameToSubSections = dict[str, "AutopkgtestLogParser.SubSection"]

    _PREFIX = r"\s*\d+s\s+autopkgtest\s+\[\d\d:\d\d:\d\d\]:"
    _RE_STARTING = re.compile(rf"{_PREFIX}\s+starting date and time")
    _RE_APT_SOURCE = re.compile(rf"{_PREFIX}\s+@{{20}}\s+apt-source.*")
    _RE_TESTBED_SETUP = re.compile(rf"{_PREFIX}\s+@{{20}} test bed setup")
    _RE_TEST_PREPARING_TESTBED = re.compile(
        rf"{_PREFIX}\s+test\s+(.*):\s+preparing testbed"
    )
    _RE_ERROR = re.compile(rf"{_PREFIX}\s+ERROR:")
    _RE_TEST_STARTS = re.compile(rf"{_PREFIX}\s+test\s+(.*): ")
    _RE_TEST_RUN = re.compile(
        rf"{_PREFIX}\s+test\s+(.*): \[-----------------------"
    )
    _RE_SUMMARY = re.compile(rf"{_PREFIX}\s+@{{20}}\s+(summary)")
    _RE_RESULT_LINE = re.compile(r"^\s*\d+s\s+(\S+)\s*(PASS|FAIL|SKIP)")

    def __init__(self, log: bytes) -> None:
        """Initialize parser."""
        # splitlines() cannot be used: it's needed to separate by
        # "\n" only and splitlines() also separates by "\r" (e.g. the
        # autopkgtest logs contain a line with "\r" in the output of apt-get
        # The rstrip(b"\n") is to avoid an extra line at the end of the log
        # (splitlines() have the "keepends" parameter for this, but not split())
        self._text = log.decode(errors="replace").rstrip("\n")

    def _parse_preparation(self, lines: "peekable[tuple[int, str]]") -> Section:
        preparation = self.Section(text="Preparation", color="primary")

        for line_number, line in lines:
            # One of the two ways that the "prepare section" finishes is that
            # next line starts with a test
            (_, next_line) = lines.peek((None, ""))

            if self._RE_STARTING.match(line):
                preparation.subsections.append(
                    self.SubSection(
                        text="start run", line=line_number, expanded=True
                    )
                )
            elif self._RE_TESTBED_SETUP.match(line):
                preparation.subsections.append(
                    self.SubSection(text="test bed setup", line=line_number)
                )
            elif self._RE_APT_SOURCE.match(line):
                preparation.subsections.append(
                    self.SubSection(text="apt-source", line=line_number)
                )
                # there will not be more subsections for preparation
                break
            elif self._RE_ERROR.match(line):
                if preparation.subsections:
                    preparation.subsections[-1].color = "danger"
                    preparation.subsections[-1].expanded = True

            elif self._RE_TEST_STARTS.match(next_line):
                # preparation has finished
                break

        return preparation

    def _parse_tests(
        self, lines: "peekable[tuple[int, str]]"
    ) -> tuple[Section, TestNameToSubSections]:
        sections = self.Section(text="Tests", color=None)
        test_name_to_subsections = {}

        current_test_name = None

        for line_number, line in lines:
            if m := self._RE_TEST_PREPARING_TESTBED.match(line):
                sections.color = "primary"
                current_test_name = m.group(1)

                # include_in_table_of_contents: initially False (only for
                # the log section). But if the preparation fails,
                # it will be changed later on
                sub_section = self.SubSection(
                    text=f"preparing testbed {current_test_name}",
                    line=line_number,
                    include_in_table_of_contents=False,
                )
                sections.subsections.append(sub_section)

                test_name_to_subsections[current_test_name] = sub_section

            elif (
                self._RE_TEST_RUN.match(line) and current_test_name is not None
            ):
                sub_section = self.SubSection(
                    text=current_test_name,
                    line=line_number,
                    include_in_table_of_contents=True,
                )

                sections.subsections.append(sub_section)

                # The "preparing testbed {current_test_name}" not needed
                # to update it anymore
                test_name_to_subsections[current_test_name] = sub_section

            # The only way to know that we finished with the tests
            # is to check if next line is the "Summary"
            (_, line) = lines.peek((None, ""))
            if self._RE_SUMMARY.match(line):
                break

        return sections, test_name_to_subsections

    def _parse_closing(
        self, lines: "peekable[tuple[int, str]]"
    ) -> tuple[Section, dict[str, Result]]:
        closing = self.Section(text="Closing", color=None)
        test_name_to_result: dict[str, AutopkgtestLogParser.Result] = {}

        for line_number, line in lines:
            if self._RE_SUMMARY.match(line):
                closing.color = "primary"
                closing.subsections.append(
                    self.SubSection(
                        text="summary", line=line_number, expanded=True
                    )
                )

            if (m := self._RE_RESULT_LINE.match(line)) is not None:
                test_name = m.group(1)
                assert isinstance(test_name, str)
                test_name_to_result[test_name] = self.Result(m.group(2))

        return closing, test_name_to_result

    def parse(self) -> tuple[list[Section], int]:
        """Parse log file. Return (list[Section], total_lines)."""
        lines = self._text.split("\n")
        total_lines = len(lines)
        lines_peekable = peekable(enumerate(lines, start=1))

        preparation = self._parse_preparation(lines_peekable)
        sections, test_name_to_subsections = self._parse_tests(lines_peekable)
        closing, results = self._parse_closing(lines_peekable)

        Result = self.Result
        for test_name, test_result in results.items():
            subsection = test_name_to_subsections.get(test_name)

            if subsection is None:
                # In case a log file is inconsistent
                continue

            match test_result:
                case Result.PASS:
                    subsection.color = "success"
                    subsection.tooltip = "Pass"
                case Result.SKIP:
                    subsection.color = "secondary"
                    subsection.tooltip = "Skip"
                case Result.FAIL:
                    subsection.color = "danger"
                    subsection.tooltip = "Fail"

                    # The relevant subsection might be the
                    # "preparing testbed" which was not in the
                    # table of contents. Adds it now
                    subsection.expanded = True
                    subsection.include_in_table_of_contents = True
                case _ as unreachable:
                    assert_never(unreachable)

        return [preparation, sections, closing], total_lines


class AutopkgtestArtifactPlugin(ArtifactPlugin):
    """Plugin for AutopkgtestArtifact."""

    artifact_category = ArtifactCategory.AUTOPKGTEST
    template_name = "web/artifact-detail.html"
    name = "autopkgtest"

    object_name = "artifact"

    @override
    def get_context_data(self) -> dict[str, Any]:
        slug = "autopkgtest-log"

        specialized_tab = {
            "specialized_tab": {
                "label": "Autopkgtest log",
                "slug": slug,
                "template": "web/_autopkgtest-artifact-detail.html",
            }
        }

        try:
            log_file_in_artifact = self.artifact.fileinartifact_set.get(
                path="log"
            )
        except FileInArtifact.DoesNotExist:
            return {
                **specialized_tab,
                "autopkgtest_artifact_error": "Artifact does not contain log",
            }

        try:
            autopkgtest_contents = view_utils.read_file(
                log_file_in_artifact, max_file_size=50 * 1024 * 1024
            )
        except view_utils.FileInArtifactError as exc:
            return {
                **specialized_tab,
                "autopkgtest_artifact_error": exc.to_html(),
            }

        sections, total_lines = AutopkgtestLogParser(
            autopkgtest_contents
        ).parse()

        file_sections = []

        subsections = list(
            itertools.chain.from_iterable(
                log_section.subsections for log_section in sections
            )
        )

        for subsection, subsection_next in itertools.pairwise(subsections):
            file_sections.append(
                LogFileWidget.Section(
                    title=subsection.text,
                    start_line=subsection.line,
                    end_line=subsection_next.line - 1,
                    expanded=subsection.expanded,
                )
            )

        # Add the latest remaining section
        if subsections:
            last_subsection = subsections[-1]
            file_sections.append(
                LogFileWidget.Section(
                    title=last_subsection.text,
                    start_line=last_subsection.line,
                    end_line=total_lines,
                    expanded=last_subsection.expanded,
                )
            )

        widget = FileWidget.create(
            file_in_artifact=log_file_in_artifact,
            file_contents=autopkgtest_contents,
            sections=file_sections,
        )

        return {
            **specialized_tab,
            "sections": sections,
            "log_widget": widget,
        }



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command file_store."""

import io

import yaml
from django.core.management import CommandError
from django.test import override_settings

from debusine.assets.models import (
    AWSProviderAccountConfiguration,
    AWSProviderAccountCredentials,
    AWSProviderAccountData,
)
from debusine.db.models import File, FileStore, Scope
from debusine.django.management.tests import call_command
from debusine.server.management.commands.file_store import Command
from debusine.test.django import TestCase


@override_settings(LANGUAGE_CODE="en-us")
class FileStoreCommandTests(TestCase):
    """Tests for the `file_store` command."""

    def test_create_from_file(self) -> None:
        """`file_store create` creates a new file store (config in file)."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        configuration_file = self.create_temporary_file(
            contents=yaml.safe_dump(configuration).encode()
        )
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            name,
            backend,
            "--configuration",
            str(configuration_file),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)
        self.assertTrue(file_store.instance_wide)
        self.assertIsNone(file_store.soft_max_size)
        self.assertIsNone(file_store.max_size)
        self.assertIsNone(file_store.provider_account)

    def test_create_from_stdin(self) -> None:
        """`file_store create` creates a new file store (config in stdin)."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)
        self.assertTrue(file_store.instance_wide)
        self.assertIsNone(file_store.soft_max_size)
        self.assertIsNone(file_store.max_size)
        self.assertIsNone(file_store.provider_account)

    def test_create_external(self) -> None:
        """`file_store create` creates a new external file store."""
        name = "test"
        backend = FileStore.BackendChoices.EXTERNAL_DEBIAN_SUITE
        configuration = {
            "archive_root_url": "https://deb.debian.org/debian",
            "suite": "bookworm",
            "components": ["main"],
        }
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)
        self.assertTrue(file_store.instance_wide)
        self.assertIsNone(file_store.soft_max_size)
        self.assertIsNone(file_store.max_size)
        self.assertIsNone(file_store.provider_account)

    def test_create_non_instance_wide(self) -> None:
        """`file_store create` creates a non-instance-wide file store."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--no-instance-wide",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)
        self.assertFalse(file_store.instance_wide)
        self.assertIsNone(file_store.soft_max_size)
        self.assertIsNone(file_store.max_size)
        self.assertIsNone(file_store.provider_account)

    def test_create_max_sizes(self) -> None:
        """`file_store create` creates a file store with maximum sizes."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--soft-max-size",
            "1024",
            "--max-size",
            "2048",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)
        self.assertTrue(file_store.instance_wide)
        self.assertEqual(file_store.soft_max_size, 1024)
        self.assertEqual(file_store.max_size, 2048)
        self.assertIsNone(file_store.provider_account)

    def test_create_provider_account(self) -> None:
        """`file_store create` creates a file store with a cloud provider."""
        asset = self.playground.create_cloud_provider_account_asset()
        name = "test"
        backend = FileStore.BackendChoices.S3
        configuration = {"bucket_name": "test-bucket"}
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--provider-account",
            asset.data["name"],
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)
        self.assertTrue(file_store.instance_wide)
        self.assertIsNone(file_store.soft_max_size)
        self.assertIsNone(file_store.max_size)
        self.assertEqual(file_store.provider_account, asset)

    def test_create_nonexistent_provider_account(self) -> None:
        """`file_store create` returns error: provider account not found."""
        self.playground.create_cloud_provider_account_asset()
        with self.assertRaisesRegex(
            CommandError,
            r"^Cloud provider account asset 'nonexistent' not found",
        ) as exc:
            call_command(
                "file_store",
                "create",
                "--provider-account",
                "nonexistent",
                "test",
                FileStore.BackendChoices.S3,
                stdin=io.StringIO(
                    yaml.safe_dump({"bucket_name": "test-bucket"})
                ),
            )

        self.assertEqual(exc.exception.returncode, 3)
        self.assertFalse(FileStore.objects.filter(name="test").exists())

    def test_create_invalid_data_yaml(self) -> None:
        """`file_store create` returns error: cannot parse YAML data."""
        with self.assertRaisesRegex(
            CommandError, r"^Error parsing YAML:"
        ) as exc:
            call_command(
                "file_store",
                "create",
                "test",
                FileStore.BackendChoices.LOCAL,
                stdin=io.StringIO(":"),
            )

        self.assertEqual(exc.exception.returncode, 3)
        self.assertFalse(FileStore.objects.filter(name="test").exists())

    def test_create_invalid_data(self) -> None:
        """`file_store create` returns error: data is invalid."""
        with self.assertRaisesRegex(
            CommandError, r"^Error creating file store:"
        ) as exc:
            call_command(
                "file_store",
                "create",
                "test",
                FileStore.BackendChoices.EXTERNAL_DEBIAN_SUITE,
                stdin=io.StringIO(yaml.safe_dump({})),
            )

        self.assertEqual(exc.exception.returncode, 3)
        self.assertFalse(FileStore.objects.filter(name="test").exists())

    def test_create_idempotent(self) -> None:
        """`file_store create` with matching parameters succeeds."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration=configuration,
        )
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store = FileStore.objects.get(name="test")
        self.assertEqual(file_store.backend, backend)
        self.assertEqual(file_store.configuration, configuration)
        self.assertTrue(file_store.instance_wide)

    def test_update_backend(self) -> None:
        """`file_store create` refuses to change the backend."""
        FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration={},
        )
        with self.assertRaisesRegex(
            CommandError, r"^Cannot change backend of existing file store$"
        ) as exc:
            call_command(
                "file_store",
                "create",
                "test",
                FileStore.BackendChoices.EXTERNAL_DEBIAN_SUITE,
                stdin=io.StringIO(
                    yaml.safe_dump(
                        {
                            "archive_root_url": "https://deb.debian.org/debian",
                            "suite": "bookworm",
                            "components": ["main"],
                        }
                    )
                ),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_update_configuration(self) -> None:
        """`file_store create` refuses to change the backend."""
        FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration={"base_directory": "/nonexistent"},
        )
        with self.assertRaisesRegex(
            CommandError,
            r"^Cannot change configuration of existing file store$",
        ) as exc:
            call_command(
                "file_store",
                "create",
                "test",
                FileStore.BackendChoices.LOCAL,
                stdin=io.StringIO(
                    yaml.safe_dump({"base_directory": "/nonexistent2"})
                ),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_update_instance_wide(self) -> None:
        """`file_store create` can change `instance_wide`."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        file_store = FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration=configuration,
        )
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--no-instance-wide",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertFalse(file_store.instance_wide)

        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertFalse(file_store.instance_wide)

        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--instance-wide",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertTrue(file_store.instance_wide)

    def test_update_max_sizes(self) -> None:
        """`file_store create` can change `soft_max_size` and `max_size`."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        file_store = FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration=configuration,
        )
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--soft-max-size",
            "1024",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertEqual(file_store.soft_max_size, 1024)
        self.assertIsNone(file_store.max_size)

        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--max-size",
            "2048",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertEqual(file_store.soft_max_size, 1024)
        self.assertEqual(file_store.max_size, 2048)

        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertEqual(file_store.soft_max_size, 1024)
        self.assertEqual(file_store.max_size, 2048)

    def test_update_provider_account(self) -> None:
        """`file_store create` can change `provider_account`."""
        provider_account_data_models = [
            AWSProviderAccountData(
                name=name,
                configuration=AWSProviderAccountConfiguration(
                    region_name="test-region"
                ),
                credentials=AWSProviderAccountCredentials(
                    access_key_id=f"{name}-access-key",
                    secret_access_key=f"{name}-secret-key",
                ),
            )
            for name in ("account1", "account2")
        ]
        provider_accounts = [
            self.playground.create_cloud_provider_account_asset(data=data)
            for data in provider_account_data_models
        ]
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        file_store = FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration=configuration,
            provider_account=provider_accounts[0],
        )
        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            "--provider-account",
            provider_account_data_models[1].name,
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertEqual(file_store.provider_account, provider_accounts[1])

        stdout, stderr, exit_code = call_command(
            "file_store",
            "create",
            name,
            backend,
            stdin=io.StringIO(yaml.safe_dump(configuration)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        file_store.refresh_from_db()
        self.assertEqual(file_store.provider_account, provider_accounts[1])

    def test_update_instance_wide_violates_constraint(self) -> None:
        """`file_store create` reports failure to change `instance_wide`."""
        name = "test"
        backend = FileStore.BackendChoices.LOCAL
        configuration = {"base_directory": "/nonexistent"}
        file_store = FileStore.objects.create(
            name="test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration=configuration,
        )
        for scope_name in ("scope1", "scope2"):
            file_store.scopes.add(
                self.playground.get_or_create_scope(scope_name)
            )

        with self.assertRaisesRegex(
            CommandError,
            r'^Error creating file store: '
            r'duplicate key value violates unique constraint '
            r'"db_filestoreinscope_unique_file_store_not_instance_wide"',
        ):
            call_command(
                "file_store",
                "create",
                "--no-instance-wide",
                name,
                backend,
                stdin=io.StringIO(yaml.safe_dump(configuration)),
            )

    def create_memory_file_store(self, *, name: str) -> FileStore:
        """Create a MEMORY file store."""
        return FileStore.objects.create(
            name=name,
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": name},
        )

    def test_delete_missing_file_store(self) -> None:
        """A nonexistent file store cannot be deleted."""
        with self.assertRaisesRegex(
            CommandError, r"^File store 'nonexistent' does not exist"
        ) as exc:
            call_command("file_store", "delete", "nonexistent")

        self.assertEqual(exc.exception.returncode, 3)

    def test_delete_with_files(self) -> None:
        """A file store with files cannot be deleted without --force."""
        name = "test"
        file_store = self.create_memory_file_store(name=name)
        file_backend = file_store.get_backend_object()
        for contents in (b"abc", b"def", b"ghi"):
            self.playground.create_file_in_backend(file_backend, contents)

        with self.assertRaisesRegex(
            CommandError,
            rf"File store {name!r} still contains 3 files; drain it first",
        ) as exc:
            call_command("file_store", "delete", name)

        self.assertEqual(exc.exception.returncode, 3)

    def test_delete_with_files_and_force(self) -> None:
        """A file store with files can be deleted with --force."""
        name = "test"
        file_store = self.create_memory_file_store(name=name)
        file_backend = file_store.get_backend_object()
        files = [
            self.playground.create_file_in_backend(file_backend, contents)
            for contents in (b"abc", b"def", b"ghi")
        ]

        call_command("file_store", "delete", "--force", name)
        self.assertQuerySetEqual(FileStore.objects.filter(name=name), [])
        # The previously-related files still exist.
        self.assertQuerySetEqual(
            File.objects.filter(pk__in=[file.pk for file in files]),
            files,
            ordered=False,
        )

    def test_delete_with_scopes(self) -> None:
        """A file store with scopes cannot be deleted without --force."""
        name = "test"
        file_store = self.create_memory_file_store(name=name)
        for scope_name in ("scope1", "scope2"):
            file_store.scopes.add(
                self.playground.get_or_create_scope(scope_name)
            )

        with self.assertRaisesRegex(
            CommandError,
            rf"File store {name!r} is still used by 2 scopes; use "
            rf"'debusine-admin scope remove_file_store' first",
        ) as exc:
            call_command("file_store", "delete", name)

        self.assertEqual(exc.exception.returncode, 3)

    def test_delete_with_scopes_and_force(self) -> None:
        """A file store with scopes can be deleted with --force."""
        name = "test"
        file_store = self.create_memory_file_store(name=name)
        for scope_name in ("scope1", "scope2"):
            file_store.scopes.add(
                self.playground.get_or_create_scope(scope_name)
            )

        call_command("file_store", "delete", "--force", name)
        self.assertQuerySetEqual(FileStore.objects.filter(name=name), [])
        # The previously-related scopes still exist.
        self.assertEqual(
            Scope.objects.filter(name__in={"scope1", "scope2"}).count(), 2
        )

    def test_delete_without_scopes(self) -> None:
        """A file store without scopes can be deleted without --force."""
        name = "test"
        self.create_memory_file_store(name=name)

        call_command("file_store", "delete", name)
        self.assertQuerySetEqual(FileStore.objects.filter(name=name), [])

    def test_unexpected_action(self) -> None:
        """Test a subcommand with no implementation."""
        command = Command()

        with self.assertRaisesRegex(
            CommandError, r"Action 'does_not_exist' not found"
        ) as exc:
            command.handle(action="does_not_exist")

        self.assertEqual(getattr(exc.exception, "returncode"), 3)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test for the main entry point of debusine."""

import base64
import datetime as dt
import secrets
from contextlib import ExitStack
from typing import Any, NotRequired, TypedDict, override

import pydantic

from debusine.assets import AssetCategory, KeyPurpose, SigningKeyData
from debusine.client.models import (
    ArtifactCreateRequest,
    AssetResponse,
    AssetsResponse,
    CollectionItemResponse,
    CollectionRoles,
    CollectionTokenGrantRequest,
    CreateWorkflowRequest,
    EnrollConfirmPayload,
    EnrollOutcome,
    EnrollPayload,
    FileRequest,
    FilesRequestType,
    Link,
    LookupMultipleResponse,
    LookupResultType,
    LookupSingleResponse,
    LookupSingleResponseArtifact,
    LookupSingleResponseCollection,
    RelationCreateRequest,
    RelationResponse,
    RelationType,
    RelationsResponse,
    ServerObject,
    StrictBaseModel,
    WorkspaceInheritanceChain,
    WorkspaceInheritanceChainElement,
)
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_work_request_response,
)


class LinkTests(TestCase):
    """Tests for :py:class:`Link`."""

    def test_can_be_str(self) -> None:
        for link, expected in (
            (Link(href="foo"), True),
            (Link(href="foo", type="bar"), False),
        ):
            with self.subTest(link=link.model_dump(exclude_unset=True)):
                self.assertEqual(link.can_be_str(), expected)

    def test_extra_arguments(self) -> None:
        link = Link(href="foo", rel="bar", title="Baz", type="wibble")
        self.assertEqual(link.href, "foo")
        self.assertEqual(link.type, "wibble")
        self.assertFalse(hasattr(link, "rel"))


class ServerObjectTests(TestCase):
    """Tests for :py:class:`ServerObject`."""

    def test_extra_json_fields(self) -> None:
        # Ensure client models ignore extra fields in serialized data, so that
        # new ones can be added without breaking old clients
        o = ServerObject(extra=1)
        self.assertEqual(o.links, {})
        self.assertFalse(hasattr(o, "extra"))

    def test_get_link_url(self) -> None:
        o = ServerObject(
            links={"simple": "/a", "complex": Link(href="/b", type="complex")}
        )
        self.assertEqual(o.get_link_url("simple"), "/a")
        self.assertEqual(o.get_link_url("complex"), "/b")


class WorkRequestTests(TestCase):
    """Tests for WorkRequest."""

    def test_work_request__str__(self) -> None:
        """__str__() return WorkRequest: {_id}."""
        work_request_id = 5
        work_request = create_work_request_response(id=work_request_id)

        self.assertEqual(str(work_request), f'WorkRequest: {work_request_id}')


class CreateWorkflowRequestTests(TestCase):
    """Tests for :py:class:`CreateWorkflowRequest`."""

    def test_validate_names(self) -> None:
        for data, expected_ok in (
            ({}, False),
            ({"template_name": "noop-template"}, True),
            ({"task_name": "noop"}, True),
            ({"template_name": "noop-template", "task_name": "noop"}, False),
        ):
            with self.subTest(data=str(data)), ExitStack() as stack:
                if not expected_ok:
                    stack.enter_context(
                        self.assertRaisesRegex(
                            pydantic.ValidationError,
                            r"Exactly one of template_name and task_name must"
                            r" be set",
                        )
                    )
                CreateWorkflowRequest(**data, task_data={})


class FileTests(TestCase):
    """Tests for the FileRequest."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.size = 10
        self.checksums = {
            "sha256": (
                "bf2cb58a68f684d95a3b78ef8f661c"
                "9a4e5b09e82cc8f9cc88cce90528caeb27"
            )
        }

        self.file = FileRequest(
            size=self.size, checksums=self.checksums, type="file"
        )

    def test_fields(self) -> None:
        """Assert expected fields."""
        self.assertEqual(self.file.size, 10)
        self.assertEqual(self.file.checksums, self.checksums)
        self.assertEqual(self.file.type, "file")

    def test_size_positive_integer(self) -> None:
        """Assert size cannot be negative."""
        with self.assertRaisesRegex(
            pydantic.ValidationError,
            r"Input should be greater than or equal to 0",
        ):
            FileRequest(size=-10, checksums=self.checksums, type="file")

    def test_size_zero_is_valid(self) -> None:
        """Assert size can be zero."""
        FileRequest(size=0, checksums=self.checksums, type="file")

    def test_file_type_only_file(self) -> None:
        """Assert file_type cannot be "link" (only "file")."""
        with self.assertRaisesRegex(
            pydantic.ValidationError,
            r"Input should be 'file'.*input_value='link'",
        ):
            FileRequest(size=self.size, checksums=self.checksums, type="link")

    def test_checksum_not_bigger_than_255(self) -> None:
        """Assert checksums cannot be longer than 255."""
        checksums = {"sha256": "a" * 256}
        with self.assertRaisesRegex(
            pydantic.ValidationError,
            r"String should have at most 255 characters",
        ):
            FileRequest(size=self.size, checksums=checksums, type="file")


class ArtifactCreateRequestTests(TestCase):
    """Tests for the ArtifactCreateRequest."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.category = "debian"
        self.workspace = "debian-lts"
        self.files = FilesRequestType(
            {
                "README": FileRequest(
                    size=10, checksums={"sha256": "aaaa"}, type="file"
                )
            }
        )
        self.data: dict[str, Any] = {}
        self.expire_at = dt.datetime.now(tz=dt.UTC) + dt.timedelta(days=1)

        self.artifact = ArtifactCreateRequest(
            category=self.category,
            workspace=self.workspace,
            files=self.files,
            data=self.data,
            expire_at=self.expire_at,
        )

    def test_fields(self) -> None:
        """Assert expected fields are accessible."""
        self.assertEqual(self.artifact.category, self.category)
        self.assertEqual(self.artifact.workspace, self.workspace)
        self.assertEqual(self.artifact.files, self.files)
        self.assertEqual(self.artifact.data, self.data)
        self.assertEqual(self.artifact.expire_at, self.expire_at)

    def test_valid_file(self) -> None:
        """Assert files must contain a file type."""
        files = FilesRequestType({"README": "test"})  # type: ignore[dict-item]
        with self.assertRaises(pydantic.ValidationError):
            self.artifact = ArtifactCreateRequest(
                category=self.category,
                workspace=self.workspace,
                files=files,
                data=self.data,
            )

    def test_create(self) -> None:
        """Assert create() returns the correct model."""
        contents = b"This is a test"

        file = self.create_temporary_file(contents=contents)

        file_request = FileRequest.create_from(file)
        self.assertEqual(file_request.size, len(contents))


class ArtifactResponseTests(TestCase):
    """Tests for ArtifactResponse."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.artifact_id = 5
        self.files_to_upload = ["README", "src/.dirstamp"]

        self.artifact = create_artifact_response(
            id=self.artifact_id,
            category="Testing",
            workspace="some workspace",
            created_at=dt.datetime.now(dt.UTC),
            download_tar_gz_url="https://example.com/some/path/",
            files_to_upload=self.files_to_upload,
        )

    def test_fields(self) -> None:
        """Test fields of ArtifactResponse."""
        self.assertEqual(self.artifact.id, self.artifact_id)
        self.assertEqual(self.artifact.files_to_upload, self.files_to_upload)


class RelationCreateRequestTests(TestCase):
    """Tests for RelationCreateRequest."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.artifact_id = 1
        self.target_id = 2

        self.relation = RelationCreateRequest(
            artifact=self.artifact_id,
            target=self.target_id,
            type=RelationType.EXTENDS,
        )

    def test_create(self) -> None:
        """Test object has the expected fields."""
        self.assertEqual(self.relation.artifact, self.artifact_id)
        self.assertEqual(self.relation.target, self.target_id)
        self.assertEqual(self.relation.type, "extends")

    def test_relation_types(self) -> None:
        """No exception is raised when setting valid relations."""
        for relation_type in RelationType:
            RelationCreateRequest(
                artifact=self.artifact_id,
                target=self.target_id,
                type=relation_type,
            )

    def test_relation_type_raise_error(self) -> None:
        """Exception is raised if a non-valid relation type is set."""
        with self.assertRaises(pydantic.ValidationError):
            RelationCreateRequest(artifact=1, target=2, type="does-not-exist")


class RelationsResponseTests(TestCase):
    """Tests for RelationsResponse."""

    def test_iter(self) -> None:
        """`__iter__` iterates over individual relation responses."""
        response = RelationsResponse.model_validate(
            [
                {"id": i, "artifact": i + 1, "target": i, "type": "extends"}
                for i in (1, 2)
            ]
        )

        self.assertEqual(
            list(response),
            [
                RelationResponse(
                    id=i, artifact=i + 1, target=i, type=RelationType.EXTENDS
                )
                for i in (1, 2)
            ],
        )


class LookupSingleResponseTests(TestCase):
    """Tests for LookupSingleResponse."""

    def test_subclass_eq(self) -> None:
        self.assertEqual(
            LookupSingleResponse(
                result_type=LookupResultType.ARTIFACT, artifact=12
            ),
            LookupSingleResponseArtifact(artifact=12),
        )
        self.assertEqual(
            LookupSingleResponse(
                result_type=LookupResultType.COLLECTION, collection=1
            ),
            LookupSingleResponseCollection(collection=1),
        )

    def test_subclass_neq(self) -> None:
        self.assertNotEqual(
            LookupSingleResponse(
                result_type=LookupResultType.ARTIFACT, artifact=12
            ),
            LookupSingleResponseArtifact(artifact=13),
        )
        self.assertNotEqual(
            LookupSingleResponse(
                result_type=LookupResultType.ARTIFACT,
                artifact=12,
                parent_collection_lookup="something",
            ),
            LookupSingleResponseArtifact(artifact=12),
        )
        self.assertNotEqual(
            LookupSingleResponse(
                result_type=LookupResultType.ARTIFACT,
                artifact=12,
                collection_item=CollectionItemResponse(
                    name="something", data={}
                ),
            ),
            LookupSingleResponseArtifact(artifact=12),
        )

    def test_unrelated_neq(self) -> None:
        self.assertNotEqual(LookupSingleResponseArtifact(artifact=1), 1)


class LookupMultipleResponseTests(TestCase):
    """Tests for LookupMultipleResponse."""

    def test_iter(self) -> None:
        """`__iter__` iterates over individual lookup results."""
        response: LookupMultipleResponse[LookupSingleResponseArtifact] = (
            LookupMultipleResponse.model_validate(
                [
                    {"result_type": "artifact", "artifact": artifact}
                    for artifact in (1, 2)
                ]
            )
        )

        self.assertEqual(
            list(response),
            [
                LookupSingleResponseArtifact(artifact=artifact)
                for artifact in (1, 2)
            ],
        )


class StrictBaseClass(StrictBaseModel):
    """Class to instantiate in StrictBaseModelTests."""

    id: int


class StrictBaseModelTests(TestCase):
    """Tests for StrictBaseModel."""

    def test_invalid_field_content_assignment(self) -> None:
        """Raise ValidationError: assign a string to an int field."""
        id_ = 5
        base_class = StrictBaseClass(id=id_)
        self.assertEqual(base_class.id, id_)

        with self.assertRaises(pydantic.ValidationError):
            base_class.id = "invalid-value"  # type: ignore[assignment]

    def test_invalid_field_in_init(self) -> None:
        """Raise ValidationError: try to initialize a non-existing field."""
        with self.assertRaises(pydantic.ValidationError):
            StrictBaseClass(does_not_exist=20)  # type: ignore[call-arg]


class AssetResponseTests(TestCase):
    """Tests for the AssetResponse model."""

    def test_parse(self) -> None:
        """Test parsing a nested dictionary."""
        body = {
            "id": 99,
            "category": "debusine:signing-key",
            "workspace": "test",
            "data": {
                "purpose": "openpgp",
                "fingerprint": "ABC123",
                "public_key": base64.b64encode(b"PUBLIC KEY").decode(),
                "description": "Description",
            },
            "work_request": 5,
        }
        AssetResponse(**body)

    def test_parse_no_category(self) -> None:
        """Test parsing an asset without a category."""
        body = {
            "id": 99,
            "category": "",
            "workspace": "test",
            "data": {},
            "work_request": 5,
        }
        with self.assertRaisesRegex(pydantic.ValidationError, r".*category.*"):
            AssetResponse(**body)


class AssetsResponsesTests(TestCase):
    """Tests for the AssetsResponses model."""

    def test_iter(self) -> None:
        """Test that AssetsResponses are iterable."""
        responses = AssetsResponse(
            [
                AssetResponse(
                    id=99,
                    category=AssetCategory.SIGNING_KEY,
                    workspace="test",
                    data=SigningKeyData(
                        purpose=KeyPurpose.OPENPGP_UPLOAD,
                        fingerprint="ABC123",
                        public_key=base64.b64encode(b"PUBLIC KEY").decode(),
                        description="Description",
                    ).model_dump(),
                )
            ]
        )
        self.assertEqual(len(list(responses)), 1)


class EnrollPayloadTests(TestCase):
    """Test for :py:class:`EnrollPayload`."""

    def test_invalid_nonces(self) -> None:
        """Invalid nonces are caught."""
        for nonce in (
            # Too short
            "",
            "a",
            "a" * 7,
            # Too long
            "a" * 65,
            "a" * 128,
            # Path components
            "/foo/bar",
            "../foobar",
            # Query strings
            "?mischief",
            # Phrases
            "foo bar baz",
        ):
            with (
                self.subTest(nonce=nonce),
                self.assertRaisesRegex(
                    ValueError, r"\bnonce\s+String should match pattern"
                ),
            ):
                EnrollPayload(
                    nonce=nonce,
                    challenge="correct horse battery staple",
                    hostname="hostname",
                    scope="scope",
                )

    def test_valid_nonces(self) -> None:
        """Acceptable nonces."""
        # The character set used by secrets.token_urlsafe should be accepted
        for nonce in (
            # Between 8 and 64 characters
            "a" * 8,
            "a" * 64,
            # Letters (upper and lowercase)
            "ABCDefgh",
            # Numbers
            "12345678",
            # Underscores and dashes
            "_-_-_-_-",
        ):
            with self.subTest(nonce=nonce):
                payload = EnrollPayload(
                    nonce=nonce,
                    challenge="correct horse battery staple",
                    hostname="hostnmame",
                    scope="scope",
                )
                self.assertEqual(payload.nonce, nonce)

    def test_invalid_challenges(self) -> None:
        """Invalid challenges are caught."""
        for challenge in (
            # Empty string
            "",
            # Blank string
            " " * 16,
            "\t" * 16,
            "\n" * 16,
            # Nonwords
            "...... ..... .....",
            # Short words
            "a b c d e",
            "foo bar baz",
            # At least 3 words
            "wibble",
            "wibble wobble",
            # Maximum 8 words
            "first second third fourth fifth sixth seventh eight ninth",
            # Words of up to 10 characters
            f"{'a' * 11} {'a' * 11} {'a' * 11}",
            # Non-ASCII letters
            "ÆØÅ àéíòùý",
            # HTML
            "<blink>Don't blink</blink>",
        ):
            with (
                self.subTest(challenge=challenge),
                self.assertRaisesRegex(
                    ValueError, r"\bchallenge\s+String should match pattern"
                ),
            ):
                EnrollPayload(
                    nonce="12345678",
                    challenge=challenge,
                    hostname="hostname",
                    scope="scope",
                )

    def test_valid_challenges(self) -> None:
        """Acceptable challenges."""
        for challenge in (
            # Words of at least 4 letters
            "ciao ciao ciao",
            # At least 3 words
            "wibble wabble wobble wubble",
            # Up to 8 words
            "first second third fourth fifth sixth seventh eight",
            # Words of up to 10 characters
            f"{'a' * 8} {'a' * 9} {'a' * 10}",
            # XKCD passwords
            "correct horse battery staple",
        ):
            with self.subTest(challenge=challenge):
                payload = EnrollPayload(
                    nonce="12345678",
                    challenge=challenge,
                    hostname="hostname",
                    scope="scope",
                )
                self.assertEqual(payload.challenge, challenge)


class EnrollConfirmPayloadTests(TestCase):
    """Test for :py:class:`EnrollConfirmPayload`."""

    def test_invalid_tokens(self) -> None:
        """Invalid tokens are caught."""
        for token in (
            # Too short
            "",
            "a",
            "a" * 7,
            # Too long
            "a" * 65,
            "a" * 128,
            # Path components
            "/foo/bar",
            "../foobar",
            # Query strings
            "?mischief",
            # Phrases
            "foo bar baz",
            # Blank string
            " " * 16,
            "\t" * 16,
            "\n" * 16,
            # Only letters and numbers
            "................",
            "wibble wobble",
            "ÆØÅ àéíòùý",
            "<blink>Don't blink</blink>",
        ):
            with (
                self.subTest(token=token),
                self.assertRaisesRegex(
                    ValueError, r"\btoken\s+String should match pattern"
                ),
            ):
                EnrollConfirmPayload(outcome=EnrollOutcome.CONFIRM, token=token)

    def test_valid_tokens(self) -> None:
        """Acceptable tokens."""
        for token in (
            # Output of token_hex
            secrets.token_hex(32),
            # Between 8 and 64 characters
            "a" * 8,
            "a" * 64,
            # Letters (upper and lowercase)
            "ABCDefgh",
            # Numbers
            "12345678",
        ):
            with self.subTest(token=token):
                payload = EnrollConfirmPayload(
                    outcome=EnrollOutcome.CONFIRM, token=token
                )
                self.assertEqual(payload.token, token)


class WorkspaceInheritanceChainElementTests(TestCase):
    """Test for :py:class:`WorkspaceInheritanceChainElement`."""

    def test_from_string(self) -> None:
        for value, expected in (
            ("42", WorkspaceInheritanceChainElement(id=42)),
            ("test", WorkspaceInheritanceChainElement(workspace="test")),
            (
                "foo/bar",
                WorkspaceInheritanceChainElement(scope="foo", workspace="bar"),
            ),
        ):
            with self.subTest(value=value):
                self.assertEqual(
                    WorkspaceInheritanceChainElement.from_string(value),
                    expected,
                )

    def test_matches(self) -> None:
        El = WorkspaceInheritanceChainElement
        sample = El(id=42, scope="scope", workspace="workspace")
        for a, b, expected in (
            (sample, El(id=42), True),
            (El(id=42, scope="scope", workspace="workspace"), El(id=43), False),
            (sample, El(workspace="workspace"), True),
            (sample, El(workspace="invalid"), False),
            (sample, El(id=42, scope="scope", workspace="workspace"), True),
            (sample, El(id=43, scope="scope", workspace="workspace"), False),
            (sample, El(id=42, scope="invalid", workspace="workspace"), False),
            (sample, El(id=42, scope="scope", workspace="invalid"), False),
        ):
            with self.subTest(a=repr(a), b=repr(b)):
                self.assertEqual(a.matches(b), expected)

    def test_invalid(self) -> None:
        class Kwargs(TypedDict):
            id: NotRequired[int]
            scope: NotRequired[str]
            workspace: NotRequired[str]

        kwargs: Kwargs
        for kwargs in (Kwargs(), Kwargs(scope="scope")):
            with (
                self.subTest(kwargs=kwargs),
                self.assertRaisesRegex(
                    ValueError, r"at least id or workspace need to be set"
                ),
            ):
                WorkspaceInheritanceChainElement(**kwargs)


class WorkspaceInheritanceChainTests(TestCase):
    """Test for :py:class:`WorkspaceInheritanceChain`."""

    def test_empty(self) -> None:
        val = WorkspaceInheritanceChain()
        self.assertEqual(val, WorkspaceInheritanceChain(chain=[]))

    def test_from_strings(self) -> None:
        El = WorkspaceInheritanceChainElement

        def chain(
            *args: WorkspaceInheritanceChainElement,
        ) -> WorkspaceInheritanceChain:
            return WorkspaceInheritanceChain(chain=list(args))

        for strings, expected in (
            ([], chain()),
            (["42"], chain(El(id=42))),
            (["test"], chain(El(workspace="test"))),
            (
                ["scope/workspace"],
                chain(El(scope="scope", workspace="workspace")),
            ),
            (
                ["1", "test", "scope/workspace"],
                chain(
                    El(id=1),
                    El(workspace="test"),
                    El(scope="scope", workspace="workspace"),
                ),
            ),
        ):
            with self.subTest(strings=strings):
                self.assertEqual(
                    WorkspaceInheritanceChain.from_strings(strings), expected
                )

    def test_add(self) -> None:
        El = WorkspaceInheritanceChainElement

        def chain(
            *args: WorkspaceInheritanceChainElement,
        ) -> WorkspaceInheritanceChain:
            return WorkspaceInheritanceChain(chain=list(args))

        el1 = El(id=1)
        el2 = El(id=2)
        for a, b, expected in (
            (chain(), chain(el1, el2), chain(el1, el2)),
            (chain(el1), chain(el2), chain(el1, el2)),
            (chain(el1, el2), chain(), chain(el1, el2)),
        ):
            self.assertEqual(a + b, expected)

    def test_sub(self) -> None:
        El = WorkspaceInheritanceChainElement

        def chain(
            *args: WorkspaceInheritanceChainElement,
        ) -> WorkspaceInheritanceChain:
            return WorkspaceInheritanceChain(chain=list(args))

        el1 = El(id=1)
        el2 = El(id=2)
        el3 = El(id=3)
        for a, b, expected in (
            (chain(), chain(el1, el2), chain()),
            (chain(el1, el2), chain(el2), chain(el1)),
            (chain(el1, el2, el3), chain(), chain(el1, el2, el3)),
            (chain(el1, el2, el3), chain(el1, el3), chain(el2)),
        ):
            with self.subTest(a=repr(a), b=repr(b)):
                self.assertEqual(a - b, expected)


class CollectionTokenGrantRequestTests(TestCase):
    """Tests for :py:class:`CollectionTokenGrantRequest`."""

    def test_valid_data(self) -> None:
        for data, expected_ok in (
            ({"token_hash": "x", "role": CollectionRoles.VIEWER}, True),
            (
                {
                    "username": "test",
                    "comment": "a comment",
                    "role": CollectionRoles.VIEWER,
                },
                True,
            ),
            (
                {
                    "token_hash": "x",
                    "username": "test",
                    "comment": "a comment",
                    "role": CollectionRoles.VIEWER,
                },
                True,
            ),
            ({"role": CollectionRoles.VIEWER}, False),
        ):
            with self.subTest(data=str(data)), ExitStack() as stack:
                if not expected_ok:
                    stack.enter_context(
                        self.assertRaisesRegex(
                            pydantic.ValidationError,
                            r"At least one of token_hash and username must be"
                            r" set",
                        )
                    )
                CollectionTokenGrantRequest(**data)
